"""."""

import time as _time
from threading import Event as _Event

import numpy as _np

from mathphys.functions import get_namedtuple as _get_namedtuple

from ..namesys import SiriusPVName as _PVName

from .device import Device as _Device, Devices as _Devices, \
    ProptyDevice as _ProptyDevice
from .dcct import DCCT
from .rf import RFCav


class BunchbyBunch(_Devices):
    """BunchbyBunch Device."""

    _devices = {
        'H': 'SI-Glob:DI-BbBProc-H',
        'V': 'SI-Glob:DI-BbBProc-V',
        'L': 'SI-Glob:DI-BbBProc-L'}
    DEVICES = _get_namedtuple('Devices', *zip(*_devices.items()))

    def __init__(self, devname):
        """."""
        devname = BunchbyBunch.process_device_name(devname)
        self.dcct = DCCT(DCCT.DEVICES.SI_13C4)
        self.rfcav = RFCav(RFCav.DEVICES.SI)
        self.info = SystemInfo(devname)
        self.timing = Timing(devname)
        self.sram = Acquisition(devname, acqtype='SRAM')
        self.bram = Acquisition(devname, acqtype='BRAM')
        self.single_bunch = SingleBunch(devname)
        self.phase_track = PhaseTracking(devname)
        self.coeffs = Coefficients(devname)
        self.feedback = Feedback(devname)
        self.drive0 = Drive(devname, drive_num=0)
        self.drive1 = Drive(devname, drive_num=1)
        self.drive2 = Drive(devname, drive_num=2)
        self.bunch_clean = BunchClean(devname)
        self.fbe = FrontBackEnd()
        devs = [
            self.info, self.timing, self.sram, self.bram, self.coeffs,
            self.feedback, self.drive0, self.drive1, self.drive2,
            self.bunch_clean, self.fbe, self.dcct,
            self.rfcav, self.single_bunch, self.phase_track]

        if devname.endswith('-L'):
            self.pwr_amp1 = PwrAmpL(devname, num=0)
            self.pwr_amp2 = PwrAmpL(devname, num=1)
            devs.append(self.pwr_amp1)
            devs.append(self.pwr_amp2)
        else:
            self.pwr_amp = PwrAmpT(devname)
            devs.append(self.pwr_amp)

        super().__init__(devname, devices=devs)

    @staticmethod
    def process_device_name(devname):
        """Check whether device exist."""
        if devname in BunchbyBunch.DEVICES:
            devname = devname
        elif devname in BunchbyBunch.DEVICES._fields:
            devname = BunchbyBunch.DEVICES[
                BunchbyBunch.DEVICES._fields.index(devname)]
        else:
            raise NotImplementedError(devname)
        return _PVName(devname)

    def sweep_phase_shifter(self, values, wait=2, mon_type='mean'):
        """Sweep Servo Phase for each `value` in `values`."""
        mon_values = []
        ctrl, mon = 'FBE Out Phase', 'SRAM Mean'
        print(f'Idx: {ctrl:15s} {mon:15s}')

        if self.devname.endswith('L'):
            propty = 'FBELT_SERVO_SETPT'
        elif self.devname.endswith('H'):
            propty = 'FBELT_X_PHASE_SETPT'
        else:
            propty = 'FBELT_Y_PHASE_SETPT'

        init_val = self.fbe[propty]
        for i, val in enumerate(values):
            self.fbe[propty] = val
            self.fbe._wait(propty, val)
            _time.sleep(wait)
            if mon_type.lower() in 'mean':
                mon_val = self.sram.data_mean
            else:
                mon_val = self.sram.spec_marker1_mag
            mon_values.append(mon_val)
            print(f'{i:03d}: {val:15.6f} {_np.mean(mon_val):15.6f}')
        self.fbe[propty] = init_val
        return _np.array(mon_values)

    def sweep_adc_delay(self, values, wait=2, mon_type='mean'):
        """Sweep ADC Delay for each `value` in `values`."""
        mon_values = []
        ctrl, mon = 'ADC Delay', 'SRAM Mean'
        print(f'Idx: {ctrl:15s} {mon:15s}')

        init_val = self.timing.adc_delay
        for i, val in enumerate(values):
            self.timing.adc_delay = val
            self.timing._wait('TADC', val)
            _time.sleep(wait)
            if mon_type.lower() in 'mean':
                mon_val = self.sram.data_mean
            else:
                mon_val = self.sram.spec_marker1_mag
            mon_values.append(mon_val)
            print(f'{i:03d}: {val:15.6f} {_np.mean(mon_val):15.6f}')
        self.timing.adc_delay = init_val
        return _np.array(mon_values)

    def sweep_backend_phase(self, values, wait=2, mon_type='peak'):
        """Sweep Backend Phase for each `value` in `values`."""
        mon_values = []
        ctrl, mon = 'Backend Phase', 'Peak magnitude'
        print(f'Idx: {ctrl:15s} {mon:15s}')

        init_val = self.fbe.be_phase
        for i, val in enumerate(values):
            self.fbe.be_phase = val
            self.fbe._wait('FBE_BE_PHASE', val)
            _time.sleep(wait)
            if mon_type.lower() in 'peak':
                mon_val = self.sram.spec_marker1_mag
            else:
                mon_val = self.sram.data_rms
            mon_values.append(mon_val)
            print(f'{i:03d}: {val:15.6f} {_np.mean(mon_val):15.6f}')
        self.fbe.be_phase = init_val
        return _np.array(mon_values)

    def sweep_dac_delay(self, values, wait=2, mon_type='peak'):
        """Sweep DAC Delay for each `value` in `values`."""
        mon_values = []
        ctrl, mon = 'DAC Delay', 'Peak Magnitude'
        print(f'Idx: {ctrl:15s} {mon:15s}')

        init_val = self.timing.dac_delay
        for i, val in enumerate(values):
            self.timing.dac_delay = val
            self.timing._wait('TDAC', val)
            _time.sleep(wait)
            if mon_type.lower() in 'peak':
                mon_val = self.sram.spec_marker1_mag
            else:
                mon_val = self.sram.data_rms
            mon_values.append(mon_val)
            print(f'{i:03d}: {val:15.6f} {_np.mean(mon_val):15.6f}')
        self.timing.dac_delay = init_val
        return _np.array(mon_values)

    def sweep_feedback_phase(self, values, wait=2, mon_type='peak'):
        """Sweep Feedback Phase for each `value` in `values`."""
        mon_values = []
        ctrl, mon = 'Coeff. Phase', 'Peak Magnitude'
        print(f'Idx: {ctrl:15s} {mon:15s}')

        init_val = self.coeffs.edit_phase
        for i, val in enumerate(values):
            self.coeffs.edit_phase = val
            self.coeffs.cmd_edit_apply()
            _time.sleep(wait)
            if mon_type.lower() in 'peak':
                mon_val = self.sram.spec_marker1_mag
            else:
                mon_val = self.sram.data_rms
            mon_values.append(mon_val)
            print(f'{i:03d}: {val:15.6f} {_np.mean(mon_val):15.6f}')
        self.coeffs.edit_phase = init_val
        self.coeffs.cmd_edit_apply()
        return _np.array(mon_values)

    def sweep_rf_phase(self, values, wait=2, mon_type='mean'):
        """Sweep RF Phase for each `value` in `values`."""
        mon_values = []
        ctrl, mon = 'RF Phase', 'SRAM Mean'
        print(f'Idx: {ctrl:15s} {mon:15s}')

        llrf = self.rfcav.dev_llrf
        init_val = llrf.phase
        for i, val in enumerate(values):
            self.rfcav.set_phase(val)
            _time.sleep(wait)
            if mon_type.lower() in 'mean':
                mon_val = self.sram.data_mean
            else:
                mon_val = self.sram.spec_marker1_mag
            mon_values.append(mon_val)
            print(f'{i:03d}: {val:15.6f} {_np.mean(mon_val):15.6f}')
        llrf.value = init_val
        return _np.array(mon_values)


class SystemInfo(_Device):
    """."""

    DEF_TIMEOUT = 10  # [s]

    _properties = (
        'ERRSUM', 'CLKMISS', 'CLKMISS_COUNT', 'PLL_UNLOCK',
        'PLL_UNLOCK_COUNT', 'DCM_UNLOCK', 'DCM_UNLOCK_COUNT', 'ADC_OVR',
        'ADC_OVR_COUNT', 'SAT', 'SAT_COUNT', 'FID_ERR', 'FID_ERR_COUNT',
        'RST_COUNT', 'CNTRST', 'RF_FREQ', 'FREV', 'HARM_NUM', 'REVISION',
        'GW_TYPE', 'IP_ADDR')

    def __init__(self, devname):
        """."""
        devname = BunchbyBunch.process_device_name(devname)

        # call base class constructor
        super().__init__(devname, properties=SystemInfo._properties)

    @property
    def status(self):
        """."""
        return self['ERRSUM']

    @property
    def clock_miss(self):
        """."""
        return self['CLKMISS']

    @property
    def clock_miss_count(self):
        """."""
        return self['CLKMISS_COUNT']

    @property
    def pll_unlock(self):
        """."""
        return self['PLL_UNLOCK']

    @property
    def pll_unlock_count(self):
        """."""
        return self['PLL_UNLOCK_COUNT']

    @property
    def dcm_unlock(self):
        """."""
        return self['DCM_UNLOCK']

    @property
    def dcm_unlock_count(self):
        """."""
        return self['DCM_UNLOCK_COUNT']

    @property
    def adc_overrange(self):
        """."""
        return self['ADC_OVR']

    @property
    def adc_overrange_count(self):
        """."""
        return self['ADC_OVR_COUNT']

    @property
    def output_saturated(self):
        """."""
        return self['SAT']

    @property
    def output_saturated_count(self):
        """."""
        return self['SAT_COUNT']

    @property
    def fiducial_error(self):
        """."""
        return self['FID_ERR']

    @property
    def fiducial_error_count(self):
        """."""
        return self['FID_ERR_COUNT']

    @property
    def time_since_last_reset(self):
        """."""
        return self['RST_COUNT']

    def cmd_reset_counts(self, timeout=DEF_TIMEOUT):
        """."""
        self['CNTRST'] = 1
        if not self._wait('CNTRST', 1, timeout/2):
            return False
        _time.sleep(0.2)
        self['CNTRST'] = 0
        return self._wait('CNTRST', 0, timeout/2)

    @property
    def rf_freq_nom(self):
        """."""
        return self['RF_FREQ'] * 1e6

    @property
    def revolution_freq_nom(self):
        """."""
        return self['FREV'] * 1e3

    @property
    def harmonic_number(self):
        """."""
        return self['HARM_NUM']

    @property
    def gateway_revision(self):
        """."""
        return self['REVISION']

    @property
    def gateway_type(self):
        """."""
        return self['GW_TYPE']

    @property
    def ip_address(self):
        """."""
        return self['IP_ADDR']


class Timing(_Device):
    """."""

    DEF_TIMEOUT = 10  # [s]

    _properties = (
        'TADC', 'TDAC', 'DELAY', 'OFF_FIDS', 'FID_DELAY', 'CLKRST',
        'FREQ_CNT_CH0', 'FREQ_CNT_CH1', 'FREQ_CNT_CH2', 'FREQ_CNT_CH3',
        'FREQ_CNT_CH4',
        'ECLDEL0', 'ECLDEL1', 'ECLDEL2', 'ECLDEL3',
        'ECLDEL0_SUBWR', 'ECLDEL1_SUBWR', 'ECLDEL2_SUBWR', 'ECLDEL3_SUBWR',
        )

    def __init__(self, devname):
        """."""
        devname = BunchbyBunch.process_device_name(devname)

        # call base class constructor
        super().__init__(devname, properties=Timing._properties)

    @property
    def adc_delay(self):
        """."""
        return self['TADC']

    @adc_delay.setter
    def adc_delay(self, value):
        self['TADC'] = value

    @property
    def dac_delay(self):
        """."""
        return self['TDAC']

    @dac_delay.setter
    def dac_delay(self, value):
        self['TDAC'] = value

    @property
    def output_delay(self):
        """."""
        return self['DELAY']

    @output_delay.setter
    def output_delay(self, value):
        self['DELAY'] = value

    @property
    def fiducial_offset(self):
        """."""
        return self['OFF_FIDS']

    @fiducial_offset.setter
    def fiducial_offset(self, value):
        self['OFF_FIDS'] = int(value)

    @property
    def fiducial_delay(self):
        """."""
        return self['FID_DELAY']

    @fiducial_delay.setter
    def fiducial_delay(self, value):
        self['FID_DELAY'] = value

    def cmd_reset_clock(self, timeout=DEF_TIMEOUT):
        """."""
        self['CLKRST'] = 1
        if not self._wait('CLKRST', 1, timeout/2):
            return False
        _time.sleep(0.2)
        self['CLKRST'] = 0
        return self._wait('CLKRST', 0, timeout/2)

    @property
    def adc_clock(self):
        """."""
        return self['ECLDEL0']

    @adc_clock.setter
    def adc_clock(self, value):
        self['ECLDEL0'] = int(value)

    @property
    def fiducial_clock(self):
        """."""
        return self['ECLDEL1']

    @fiducial_clock.setter
    def fiducial_clock(self, value):
        self['ECLDEL1'] = int(value)

    @property
    def fiducial(self):
        """."""
        return self['ECLDEL2']

    @fiducial.setter
    def fiducial(self, value):
        self['ECLDEL2'] = int(value)

    @property
    def dac_clock(self):
        """."""
        return self['ECLDEL3']

    @dac_clock.setter
    def dac_clock(self, value):
        self['ECLDEL3'] = int(value)

    @property
    def adc_clock_ok(self):
        """."""
        return self['ECLDEL0_SUBWR']

    @property
    def fiducial_clock_ok(self):
        """."""
        return self['ECLDEL1_SUBWR']

    @property
    def fiducial_ok(self):
        """."""
        return self['ECLDEL2_SUBWR']

    @property
    def dac_clock_ok(self):
        """."""
        return self['ECLDEL3_SUBWR']

    @property
    def freq_input_clock(self):
        """."""
        return self['FREQ_CNT_CH0']

    @property
    def freq_rf_over_2(self):
        """."""
        return self['FREQ_CNT_CH1']

    @property
    def freq_rf_over_3(self):
        """."""
        return self['FREQ_CNT_CH2']

    @property
    def freq_dac_clock(self):
        """."""
        return self['FREQ_CNT_CH3']

    @property
    def freq_rf_over_4(self):
        """."""
        return self['FREQ_CNT_CH4']


class Coefficients(_Device):
    """."""

    _properties = (
        'CSET0', 'CSET1', 'CSET2', 'CSET3', 'COEFF',
        'DESC_CSET0', 'DESC_CSET1', 'DESC_CSET2', 'DESC_CSET3', 'DESC_COEFF',
        'CVERIFY.C', 'CVERIFY.D', 'CVERIFY.G', 'CVERIFY.H',
        'LDSET', 'BO_CVERIFY', 'BO_CPCOEFF',
        'FLT_GAIN', 'FLT_FREQ', 'FLT_PHASE', 'FLT_TAPS',
        'FTF_MAG', 'FTF_PHASE', 'FTF_FREQ',
        'FTF_TUNE', 'FTF_GTUNE', 'FTF_PTUNE',
        )

    DEF_TIMEOUT = 10  # [s]
    OFF, ON = 0, 1
    FPGA_BITS = 2**15

    def __init__(self, devname):
        """."""
        devname = BunchbyBunch.process_device_name(devname)

        # call base class constructor
        super().__init__(devname, properties=Coefficients._properties)

    @property
    def set0(self):
        """."""
        return self['CSET0']/self.FPGA_BITS

    @set0.setter
    def set0(self, value):
        self['CSET0'] = _np.array(value) * self.FPGA_BITS

    @property
    def set1(self):
        """."""
        return self['CSET1'] / self.FPGA_BITS

    @set1.setter
    def set1(self, value):
        self['CSET1'] = _np.array(value) * self.FPGA_BITS

    @property
    def set2(self):
        """."""
        return self['CSET2'] / self.FPGA_BITS

    @set2.setter
    def set2(self, value):
        self['CSET2'] = _np.array(value) * self.FPGA_BITS

    @property
    def set3(self):
        """."""
        return self['CSET3'] / self.FPGA_BITS

    @set3.setter
    def set3(self, value):
        self['CSET3'] = _np.array(value) * self.FPGA_BITS

    @property
    def set0_desc(self):
        """."""
        return self['DESC_CSET0']

    @property
    def set1_desc(self):
        """."""
        return self['DESC_CSET1']

    @property
    def set2_desc(self):
        """."""
        return self['DESC_CSET2']

    @property
    def set3_desc(self):
        """."""
        return self['DESC_CSET3']

    @property
    def set0_status(self):
        """."""
        return self['CVERIFY.C']

    @property
    def set1_status(self):
        """."""
        return self['CVERIFY.D']

    @property
    def set2_status(self):
        """."""
        return self['CVERIFY.G']

    @property
    def set3_status(self):
        """."""
        return self['CVERIFY.H']

    @property
    def edit_choose_set(self):
        """."""
        return self['LDSET']

    @edit_choose_set.setter
    def edit_choose_set(self, value):
        """."""
        self['LDSET'] = value

    @property
    def edit_set(self):
        """."""
        return self['COEFF'] / self.FPGA_BITS

    @edit_set.setter
    def edit_set(self, value):
        """."""
        self['COEFF'] = _np.array(value) * self.FPGA_BITS

    @property
    def edit_set_desc(self):
        """."""
        return self['DESC_COEFF']

    @property
    def edit_set_status(self):
        """."""
        return self['BO_CVERIFY']

    @property
    def edit_ftf_magnitude(self):
        """."""
        return self['FTF_MAG']

    @property
    def edit_ftf_phase(self):
        """."""
        return self['FTF_PHASE']

    @property
    def edit_ftf_freq(self):
        """."""
        return self['FTF_FREQ']

    @property
    def edit_marker_freq(self):
        """."""
        return self['FTF_TUNE']

    @property
    def edit_marker_mag(self):
        """."""
        return self['FTF_GTUNE'][0]

    @property
    def edit_marker_phase(self):
        """."""
        return self['FTF_PTUNE'][0]

    @property
    def edit_gain(self):
        """."""
        return self['FLT_GAIN']

    @edit_gain.setter
    def edit_gain(self, value):
        """."""
        self['FLT_GAIN'] = value

    @property
    def edit_freq(self):
        """."""
        return self['FLT_FREQ']

    @edit_freq.setter
    def edit_freq(self, value):
        """."""
        self['FLT_FREQ'] = value

    @property
    def edit_phase(self):
        """."""
        return self['FLT_PHASE']

    @edit_phase.setter
    def edit_phase(self, value):
        """."""
        self['FLT_PHASE'] = value

    @property
    def edit_nrpoints(self):
        """."""
        return self['FLT_TAPS']

    @edit_nrpoints.setter
    def edit_nrpoints(self, value):
        """."""
        self['FLT_TAPS'] = value

    def cmd_edit_apply(self, timeout=DEF_TIMEOUT):
        """."""
        self['BO_CPCOEFF'] = 1
        return self._wait('BO_CPCOEFF', 1, timeout)

    def cmd_edit_verify(self, timeout=DEF_TIMEOUT):
        """."""
        self['BO_CVERIFY'] = 1
        return self._wait('BO_CVERIFY', 1, timeout)


class Acquisition(_ProptyDevice):
    """."""

    ACQTYPE = _get_namedtuple('Devices', ('SRAM', 'BRAM'))

    _properties = (
        'GDTIME', 'HOLDTIME', 'POSTTIME', 'ACQTIME',
        'REC_DS', 'POSTSEL', 'ACQ_EN', 'ACQ_SINGLE', 'SP_AVG',
        'HWTEN', 'TRIG_IN_SEL', 'ARM', 'ARM_MON', 'BR_ARM',
        'DUMP', 'RAW_SAMPLES', 'RAW', 'ACQ_TURNS', 'POST_TURNS',
        'MEAN', 'RMS', 'XSC', 'SPEC', 'MAXRMS', 'TSC', 'FREQ',
        'ACQ_MASK', 'ACQ_PATTERN',
        'SP_LOW1', 'SP_HIGH1', 'SP_SEARCH1',
        'SP_LOW2', 'SP_HIGH2', 'SP_SEARCH2',
        'PEAKFREQ1', 'PEAK1', 'PEAKTUNE1',
        'PEAKFREQ2', 'PEAK2', 'PEAKTUNE2',
        'MD_ENABLE', 'MD_SMODE', 'MD_FTUNE', 'MD_FSPAN', 'MD_MSEL',
        'MD_AVG', 'MD_SP_LOW', 'MD_SP_HIGH', 'MD_SP_SEARCH',
        'MD_MAXMODE', 'MD_MAXVAL', 'MD_PEAK', 'MD_PEAKFREQ', 'MD_PEAKTUNE',
        'MD_MODES', 'MD_SPEC',
        )

    DEF_TIMEOUT = 10  # [s]

    def __init__(self, devname, acqtype='BRAM'):
        """."""
        devname = BunchbyBunch.process_device_name(devname)
        acqtype = Acquisition.process_acquisition_type(acqtype)

        # call base class constructor
        super().__init__(
            devname, propty_prefix=acqtype+'_',
            properties=Acquisition._properties)

        pvo = self.pv_object('RAW')
        self._update_data_evt = _Event()
        pvo.add_callback(self._update_evt)

    # ########### Acquisition Config Properties ###########
    @property
    def acq_type(self):
        """."""
        return self['POSTSEL']

    @acq_type.setter
    def acq_type(self, value):
        self['POSTSEL'] = value

    @property
    def acq_enbl(self):
        """."""
        return self['ACQ_EN']

    @acq_enbl.setter
    def acq_enbl(self, value):
        self['ACQ_EN'] = value

    @property
    def acq_mode(self):
        """."""
        return self['ACQ_SINGLE']

    @acq_mode.setter
    def acq_mode(self, value):
        self['ACQ_SINGLE'] = value

    @property
    def growthtime(self):
        """."""
        return self['GDTIME']

    @growthtime.setter
    def growthtime(self, value):
        self['GDTIME'] = value

    @property
    def acqtime(self):
        """."""
        return self['ACQTIME']

    @acqtime.setter
    def acqtime(self, value):
        self['ACQTIME'] = value

    @property
    def holdtime(self):
        """."""
        return self['HOLDTIME']

    @holdtime.setter
    def holdtime(self, value):
        self['HOLDTIME'] = value

    @property
    def posttime(self):
        """."""
        return self['POSTTIME']

    @posttime.setter
    def posttime(self, value):
        self['POSTTIME'] = value

    @property
    def downsample(self):
        """."""
        return self['REC_DS']

    @downsample.setter
    def downsample(self, value):
        self['REC_DS'] = value

    @property
    def nr_averages(self):
        """."""
        return self['SP_AVG']

    @nr_averages.setter
    def nr_averages(self, value):
        self['SP_AVG'] = value

    # ########### Trigger Properties ###########
    @property
    def trigger_type(self):
        """."""
        return self['HWTEN']

    @trigger_type.setter
    def trigger_type(self, value):
        self['HWTEN'] = value

    @property
    def trigger_sel(self):
        """."""
        return self['TRIG_IN_SEL']

    @trigger_sel.setter
    def trigger_sel(self, value):
        self['TRIG_IN_SEL'] = value

    @property
    def trigger_armed(self):
        """."""
        return self['ARM_MON']

    @trigger_armed.setter
    def trigger_armed(self, value):
        self['ARM'] = value

    @property
    def trigger_rearm(self):
        """."""
        return self['BR_ARM']

    @trigger_rearm.setter
    def trigger_rearm(self, value):
        self['BR_ARM'] = value

    # ########### Data Properties ###########
    @property
    def data_size(self):
        """."""
        return self['RAW_SAMPLES']

    @data_size.setter
    def data_size(self, value):
        self['RAW_SAMPLES'] = value

    @property
    def data_raw(self):
        """."""
        data = self['RAW']
        size = self.data_size
        if size is not None:
            data = data[:size]
        return data

    @property
    def data_bunch_index(self):
        """."""
        return self['XSC']

    @property
    def data_mean(self):
        """."""
        return self['MEAN']

    @property
    def data_rms(self):
        """."""
        return self['RMS']

    @property
    def data_nrturns(self):
        """."""
        return self['ACQ_TURNS']

    @property
    def data_nrturns_posttriger(self):
        """."""
        return self['POST_TURNS']

    def cmd_data_acquire(self, timeout=DEF_TIMEOUT):
        """."""
        self.acq_enbl = 1
        return self._wait('ACQ_EN', 1, timeout=timeout)

    def cmd_data_dump(self, timeout=DEF_TIMEOUT, pv_update=False):
        """."""
        if pv_update:
            self._update_data_evt.clear()
        self['DUMP'] = 1
        return self.wait_data_dump(timeout, pv_update=pv_update)

    def wait_data_dump(self, timeout=None, pv_update=False):
        """."""
        timeout = timeout or Acquisition.DEF_TIMEOUT
        if not self._wait('DUMP', False, timeout=timeout):
            print('WARN: Timed out waiting data dump.')
            return False
        if pv_update and not self._update_data_evt.wait(timeout=timeout):
            print('WARN: Timed out waiting PV data update.')
            return False
        self._update_data_evt.clear()
        return True

    # ########### Spectrometer Properties ###########
    @property
    def spec_mag(self):
        """."""
        return self['SPEC']

    @property
    def spec_freq(self):
        """."""
        return self['FREQ']

    @property
    def spec_data(self):
        """."""
        return self['MAXRMS']

    @property
    def spec_time(self):
        """."""
        return self['TSC']

    @property
    def spec_mask(self):
        """."""
        return self['ACQ_MASK']

    @spec_mask.setter
    def spec_mask(self, value):
        self['ACQ_MASK'] = _np.array(value)

    @property
    def spec_mask_pattern(self):
        """."""
        return self['ACQ_PATTERN']

    @spec_mask_pattern.setter
    def spec_mask_pattern(self, value):
        self['ACQ_PATTERN'] = str(value)

    @property
    def spec_marker1_freq_min(self):
        """."""
        return self['SP_LOW1']

    @spec_marker1_freq_min.setter
    def spec_marker1_freq_min(self, value):
        self['SP_LOW1'] = value

    @property
    def spec_marker1_freq_max(self):
        """."""
        return self['SP_HIGH1']

    @spec_marker1_freq_max.setter
    def spec_marker1_freq_max(self, value):
        self['SP_HIGH1'] = value

    @property
    def spec_marker1_search_mode(self):
        """."""
        return self['SP_SEARCH1']

    @spec_marker1_search_mode.setter
    def spec_marker1_search_mode(self, value):
        self['SP_SEARCH1'] = value

    @property
    def spec_marker1_freq(self):
        """."""
        return self['PEAKFREQ1']

    @property
    def spec_marker1_tune(self):
        """."""
        return self['PEAKTUNE1']

    @property
    def spec_marker1_mag(self):
        """."""
        return self['PEAK1']

    @property
    def spec_marker2_freq_min(self):
        """."""
        return self['SP_LOW2']

    @spec_marker2_freq_min.setter
    def spec_marker2_freq_min(self, value):
        self['SP_LOW2'] = value

    @property
    def spec_marker2_freq_max(self):
        """."""
        return self['SP_HIGH2']

    @spec_marker2_freq_max.setter
    def spec_marker2_freq_max(self, value):
        self['SP_HIGH2'] = value

    @property
    def spec_marker2_search_mode(self):
        """."""
        return self['SP_SEARCH2']

    @spec_marker2_search_mode.setter
    def spec_marker2_search_mode(self, value):
        self['SP_SEARCH2'] = value

    @property
    def spec_marker2_freq(self):
        """."""
        return self['PEAKFREQ2']

    @property
    def spec_marker2_tune(self):
        """."""
        return self['PEAKTUNE2']

    @property
    def spec_marker2_mag(self):
        """."""
        return self['PEAK2']

    @staticmethod
    def process_acquisition_type(acqtype):
        """."""
        if acqtype in Acquisition.ACQTYPE:
            acqtype = Acquisition.ACQTYPE._fields[acqtype]
        elif acqtype in Acquisition.ACQTYPE._fields:
            acqtype = acqtype
        else:
            raise NotImplementedError(acqtype)
        return acqtype

    # ############# Modal Analysis Properties #############
    @property
    def modal_acq_enbl(self):
        """."""
        return self['MD_ENABLE']

    @modal_acq_enbl.setter
    def modal_acq_enbl(self, value):
        self['MD_ENABLE'] = value

    @property
    def modal_acq_mode(self):
        """."""
        return self['MD_SMODE']

    @modal_acq_mode.setter
    def modal_acq_mode(self, value):
        self['MD_SMODE'] = value

    @property
    def modal_mode_selected(self):
        """."""
        return self['MD_MSEL']

    @modal_mode_selected.setter
    def modal_mode_selected(self, value):
        self['MD_MSEL'] = value

    @property
    def modal_nr_averages(self):
        """."""
        return self['MD_AVG']

    @modal_nr_averages.setter
    def modal_nr_averages(self, value):
        self['MD_AVG'] = value

    @property
    def modal_sideband_freq(self):
        """."""
        return self['MD_FTUNE']

    @modal_sideband_freq.setter
    def modal_sideband_freq(self, value):
        self['MD_FTUNE'] = value

    @property
    def modal_sideband_span(self):
        """."""
        return self['MD_FSPAN']

    @modal_sideband_span.setter
    def modal_sideband_span(self, value):
        self['MD_FSPAN'] = value

    @property
    def modal_marker_freq_min(self):
        """."""
        return self['MD_SP_LOW']

    @modal_marker_freq_min.setter
    def modal_marker_freq_min(self, value):
        self['MD_SP_LOW'] = value

    @property
    def modal_marker_freq_max(self):
        """."""
        return self['MD_SP_HIGH']

    @modal_marker_freq_max.setter
    def modal_marker_freq_max(self, value):
        self['MD_SP_HIGH'] = value

    @property
    def modal_marker_search_mode(self):
        """."""
        return self['MD_SP_SEARCH']

    @modal_marker_search_mode.setter
    def modal_marker_search_mode(self, value):
        self['MD_SP_SEARCH'] = value

    @property
    def modal_maximum_mode(self):
        """."""
        return self['MD_MAXMODE']

    @property
    def modal_maximum_value(self):
        """."""
        return self['MD_MAXVAL']

    @property
    def modal_marker_freq(self):
        """."""
        return self['MD_PEAKFREQ']

    @property
    def modal_marker_tune(self):
        """."""
        return self['MD_PEAKTUNE']

    @property
    def modal_marker_mag(self):
        """."""
        return self['MD_PEAK']

    @property
    def modal_modes_amp(self):
        """."""
        return self['MD_MODES']

    @property
    def modal_maximum_mode_spec(self):
        """."""
        return self['MD_SPEC']

    def _update_evt(self, **kwargs):
        _ = kwargs
        self._update_data_evt.set()


class SingleBunch(_ProptyDevice):
    """."""

    _properties = (
        'ACQTIME', 'ACQ_SAMPLES', 'ACQ_EN', 'ACQ_SINGLE',
        'BUNCH_ID', 'RAW_BUNCH_ID',
        'EXTEN', 'TRIG_IN_SEL', 'ARM', 'ARM_MON', 'BR_ARM',
        'RAW_SAMPLES', 'TSC', 'RAW', 'FREQ', 'MAG', 'PHASE', 'TF_ENABLE',
        'NFFT', 'NOVERLAP', 'DEL_CAL', 'SP_AVG',
        'MEANVAL', 'RMSVAL', 'AMP_PP',
        'SP_LOW1', 'SP_HIGH1', 'PEAKFREQ1', 'PEAKTUNE1',
        'PEAK1', 'SP_SEARCH1', 'PHASE1',
        )

    DEF_TIMEOUT = 10  # [s]

    def __init__(self, devname):
        """."""
        devname = BunchbyBunch.process_device_name(devname)

        # call base class constructor
        super().__init__(
            devname, propty_prefix='SB_', properties=SingleBunch._properties)

    @property
    def acqtime(self):
        """."""
        return self['ACQTIME']

    @acqtime.setter
    def acqtime(self, value):
        self['ACQTIME'] = value

    @property
    def acq_enbl(self):
        """."""
        return self['ACQ_EN']

    @acq_enbl.setter
    def acq_enbl(self, value):
        self['ACQ_EN'] = value

    @property
    def acq_mode(self):
        """."""
        return self['ACQ_SINGLE']

    @acq_mode.setter
    def acq_mode(self, value):
        self['ACQ_SINGLE'] = value

    @property
    def bunch_id(self):
        """."""
        return self['RAW_BUNCH_ID']

    @bunch_id.setter
    def bunch_id(self, value):
        self['BUNCH_ID'] = value

    # ########### Trigger Properties ###########
    @property
    def trigger_type(self):
        """."""
        return self['EXTEN']

    @trigger_type.setter
    def trigger_type(self, value):
        self['EXTEN'] = value

    @property
    def trigger_sel(self):
        """."""
        return self['TRIG_IN_SEL']

    @trigger_sel.setter
    def trigger_sel(self, value):
        self['TRIG_IN_SEL'] = value

    @property
    def trigger_armed(self):
        """."""
        return self['ARM_MON']

    @trigger_armed.setter
    def trigger_armed(self, value):
        self['ARM'] = value

    @property
    def trigger_rearm(self):
        """."""
        return self['BR_ARM']

    @trigger_rearm.setter
    def trigger_rearm(self, value):
        self['BR_ARM'] = value

    # ########### Data Properties ###########
    @property
    def data_size(self):
        """."""
        return self['RAW_SAMPLES']

    @data_size.setter
    def data_size(self, value):
        self['RAW_SAMPLES'] = value

    @property
    def data_raw(self):
        """."""
        data = self['RAW']
        size = self.data_size
        if size is not None:
            data = data[:size]
        return data

    @property
    def data_time(self):
        """."""
        return self['TSC']

    ##
    @property
    def fft_size(self):
        """."""
        return self['NFFT']

    @fft_size.setter
    def fft_size(self, value):
        self['NFFT'] = value

    @property
    def fft_overlap(self):
        """."""
        return self['NOVERLAP']

    @fft_overlap.setter
    def fft_overlap(self, value):
        self['NOVERLAP'] = value

    @property
    def delay_calibration(self):
        """."""
        return self['DEL_CAL']

    @delay_calibration.setter
    def delay_calibration(self, value):
        self['DEL_CAL'] = value

    @property
    def nr_averages(self):
        """."""
        return self['SP_AVG']

    @nr_averages.setter
    def nr_averages(self, value):
        self['SP_AVG'] = value

    @property
    def transfer_function_enable(self):
        """."""
        return self['TF_ENABLE']

    @transfer_function_enable.setter
    def transfer_function_enable(self, value):
        self['TF_ENABLE'] = value

    @property
    def amplitude_mean(self):
        """."""
        return self['MEANVAL']

    @amplitude_mean.setter
    def amplitude_mean(self, value):
        self['MEANVAL'] = value

    @property
    def amplitude_rms(self):
        """."""
        return self['RMSVAL']

    @amplitude_rms.setter
    def amplitude_rms(self, value):
        self['RMSVAL'] = value

    @property
    def amplitude_pp(self):
        """."""
        return self['AMP_PP']

    @amplitude_pp.setter
    def amplitude_pp(self, value):
        self['AMP_PP'] = value

    @property
    def spec_mag(self):
        """."""
        return self['MAG']

    @property
    def spec_phase(self):
        """."""
        return self['PHASE']

    @property
    def spec_freq(self):
        """."""
        return self['FREQ']

    @property
    def spec_marker1_freq_min(self):
        """."""
        return self['SP_LOW1']

    @spec_marker1_freq_min.setter
    def spec_marker1_freq_min(self, value):
        self['SP_LOW1'] = value

    @property
    def spec_marker1_freq_max(self):
        """."""
        return self['SP_HIGH1']

    @spec_marker1_freq_max.setter
    def spec_marker1_freq_max(self, value):
        self['SP_HIGH1'] = value

    @property
    def spec_marker1_search_mode(self):
        """."""
        return self['SP_SEARCH1']

    @spec_marker1_search_mode.setter
    def spec_marker1_search_mode(self, value):
        self['SP_SEARCH1'] = value

    @property
    def spec_marker1_freq(self):
        """."""
        return self['PEAKFREQ1']

    @property
    def spec_marker1_tune(self):
        """."""
        return self['PEAKTUNE1']

    @property
    def spec_marker1_mag(self):
        """."""
        return self['PEAK1']

    @property
    def spec_marker1_phase(self):
        """."""
        return self['PHASE1']

    def cmd_enable_transfer_function(self, timeout=DEF_TIMEOUT):
        """Enable transfer function."""
        self.transfer_function_enable = 1
        return self._wait('TF_ENABLE', value=1, timeout=timeout)


class PhaseTracking(_Device):
    """."""

    _properties = (
        'PHTRK_GAIN', 'PHTRK_SETPT', 'PHTRK_RANGE', 'PHTRK_DECIM',
        'PHTRK_RATE', 'PHTRK_BANDWIDTH', 'PHTRK_LOOPCTRL',
        'PHTRK_MAG', 'PHTRK_TFGAIN', 'PHTRK_SHIFT', 'PHTRK_PHASE',
        'PHTRK_ERROR', 'PHTRK_FREQ0', 'PHTRK_TUNE',
        'DRIVE2_TRACK', 'PHTRK_MOD')

    DEF_TIMEOUT = 10  # [s]

    def __init__(self, devname):
        """."""
        devname = BunchbyBunch.process_device_name(devname)

        # call base class constructor
        super().__init__(
            devname, properties=PhaseTracking._properties)

    @property
    def gain(self):
        """."""
        return self['PHTRK_GAIN']

    @gain.setter
    def gain(self, value):
        self['PHTRK_GAIN'] = value

    @property
    def setpoint(self):
        """Phase setpoint in degrees."""
        return self['PHTRK_SETPT']

    @setpoint.setter
    def setpoint(self, value):
        self['PHTRK_SETPT'] = value

    @property
    def range(self):
        """Frequency range in kHz."""
        return self['PHTRK_RANGE']

    @range.setter
    def range(self, value):
        self['PHTRK_RANGE'] = value

    @property
    def decimation(self):
        """."""
        return self['PHTRK_DECIM']

    @decimation.setter
    def decimation(self, value):
        self['PHTRK_DECIM'] = value

    @property
    def loop_state(self):
        """Loop State."""
        return self['PHTRK_LOOPCTRL']

    @loop_state.setter
    def loop_state(self, value):
        self['PHTRK_LOOPCTRL'] = value

    @property
    def use_drive2(self):
        """Loop State."""
        return self['DRIVE2_TRACK']

    @use_drive2.setter
    def use_drive2(self, value):
        self['DRIVE2_TRACK'] = value

    @property
    def time_modulation(self):
        """Loop State."""
        return self['PHTRK_MOD']

    @time_modulation.setter
    def time_modulation(self, value):
        self['PHTRK_MOD'] = value

    @property
    def rate(self):
        """."""
        return self['PHTRK_RATE']

    @property
    def bandwidth(self):
        """."""
        return self['PHTRK_BANDWIDTH']

    @property
    def magnitude(self):
        """."""
        return self['PHTRK_MAG']

    @property
    def transfer_gain(self):
        """."""
        return self['PHTRK_TFGAIN']

    @property
    def normalizing_shift(self):
        """."""
        return self['PHTRK_SHIFT']

    @property
    def phase(self):
        """Phase monitor in degrees."""
        return self['PHTRK_PHASE']

    @property
    def error(self):
        """Error in degrees."""
        return self['PHTRK_ERROR']

    @property
    def frequency(self):
        """."""
        return self['PHTRK_FREQ0']

    @property
    def tune(self):
        """."""
        return self['PHTRK_TUNE']


class FrontBackEnd(_Device):
    """."""

    _properties = (
        'FBE_Z_ATT', 'FBE_Z_PHASE', 'FBELT_SERVO_SETPT',
        'FBE_BE_ATT', 'FBE_BE_PHASE',
        'FBE_X_ATT', 'FBE_X_PHASE', 'FBELT_X_PHASE_SETPT',
        'FBE_Y_ATT', 'FBE_Y_PHASE', 'FBELT_Y_PHASE_SETPT',
        )

    FPGA_BITS = 2**15

    def __init__(self):
        """."""
        # call base class constructor
        super().__init__(
            BunchbyBunch.DEVICES.L, properties=FrontBackEnd._properties)

    @property
    def z_att(self):
        """."""
        return self['FBE_Z_ATT']

    @z_att.setter
    def z_att(self, value):
        """."""
        self['FBE_Z_ATT'] = value

    @property
    def be_att(self):
        """."""
        return self['FBE_BE_ATT']

    @be_att.setter
    def be_att(self, value):
        """."""
        self['FBE_BE_ATT'] = value

    @property
    def x_att(self):
        """."""
        return self['FBE_X_ATT']

    @x_att.setter
    def x_att(self, value):
        """."""
        self['FBE_X_ATT'] = value

    @property
    def y_att(self):
        """."""
        return self['FBE_Y_ATT']

    @y_att.setter
    def y_att(self, value):
        """."""
        self['FBE_Y_ATT'] = value

    @property
    def z_phase(self):
        """."""
        return self['FBE_Z_PHASE']

    @z_phase.setter
    def z_phase(self, value):
        """."""
        self['FBELT_SERVO_SETPT'] = value

    @property
    def be_phase(self):
        """."""
        return self['FBE_BE_PHASE']

    @be_phase.setter
    def be_phase(self, value):
        """."""
        self['FBE_BE_PHASE'] = value

    @property
    def x_phase(self):
        """."""
        return self['FBE_X_PHASE']

    @x_phase.setter
    def x_phase(self, value):
        """."""
        self['FBELT_X_PHASE_SETPT'] = value

    @property
    def y_phase(self):
        """."""
        return self['FBE_Y_PHASE']

    @y_phase.setter
    def y_phase(self, value):
        """."""
        self['FBELT_Y_PHASE_SETPT'] = value


class Feedback(_Device):
    """."""

    _properties = (
        'PROC_DS', 'FBCTRL', 'SHIFTGAIN', 'SETSEL', 'SAT_THRESHOLD',
        'FB_MASK', 'FB_PATTERN', 'CF_MASK', 'CF_PATTERN',
        'CF_PATTERN_SUB.VALB', 'GDEN',
        )

    def __init__(self, devname):
        """."""
        devname = BunchbyBunch.process_device_name(devname)

        # call base class constructor
        super().__init__(devname, properties=Feedback._properties)

    @property
    def downsample(self):
        """."""
        return self['PROC_DS']

    @downsample.setter
    def downsample(self, value):
        self['PROC_DS'] = int(value)

    @property
    def loop_state(self):
        """."""
        return self['FBCTRL']

    @loop_state.setter
    def loop_state(self, value):
        self['FBCTRL'] = int(value)

    @property
    def shift_gain(self):
        """."""
        return self['SHIFTGAIN']

    @shift_gain.setter
    def shift_gain(self, value):
        self['SHIFTGAIN'] = int(value)

    @property
    def coeff_set(self):
        """."""
        return self['SETSEL']

    @coeff_set.setter
    def coeff_set(self, value):
        self['SETSEL'] = int(value)

    @property
    def saturation_threshold(self):
        """."""
        return self['SAT_THRESHOLD']

    @saturation_threshold.setter
    def saturation_threshold(self, value):
        self['SAT_THRESHOLD'] = value

    @property
    def mask(self):
        """."""
        return self['FB_MASK']

    @mask.setter
    def mask(self, value):
        self['FB_MASK'] = _np.array(value)

    @property
    def mask_pattern(self):
        """."""
        return self['FB_PATTERN']

    @mask_pattern.setter
    def mask_pattern(self, value):
        self['FB_PATTERN'] = str(value)

    @property
    def alternate_mask(self):
        """."""
        return self['CF_MASK']

    @alternate_mask.setter
    def alternate_mask(self, value):
        self['CF_MASK'] = _np.array(value)

    @property
    def alternate_mask_pattern(self):
        """."""
        return self['CF_PATTERN']

    @alternate_mask_pattern.setter
    def alternate_mask_pattern(self, value):
        self['CF_PATTERN'] = str(value)

    @property
    def alternate_nr_bunches(self):
        """."""
        return self['CF_PATTERN_SUB.VALB']

    @property
    def alternate_inuse(self):
        """."""
        return bool(self.alternate_nr_bunches)

    @property
    def grow_damp_state(self):
        """."""
        return self['GDEN']

    @grow_damp_state.setter
    def grow_damp_state(self, value):
        self['GDEN'] = int(value)


class Drive(_ProptyDevice):
    """."""

    _properties = (
        'MOD', 'AMPL', 'WAVEFORM', 'FREQ', 'FREQ_ACT', 'SPAN', 'SPAN_ACT',
        'PERIOD', 'PERIOD_ACT', 'MASK', 'PATTERN', 'BITS',
        )

    def __init__(self, devname, drive_num=None):
        """."""
        devname = BunchbyBunch.process_device_name(devname)
        propty = 'DRIVE'
        if drive_num is not None:
            propty += str(drive_num)
        propty += '_'
        super().__init__(
            devname, propty_prefix=propty, properties=Drive._properties)

    @property
    def number_of_bits(self):
        """."""
        return self['BITS']

    @property
    def state(self):
        """."""
        return self['MOD']

    @state.setter
    def state(self, value):
        self['MOD'] = int(value)

    @property
    def waveform(self):
        """."""
        return self['WAVEFORM']

    @waveform.setter
    def waveform(self, value):
        self['WAVEFORM'] = int(value)

    @property
    def amplitude(self):
        """."""
        return self['AMPL']

    @amplitude.setter
    def amplitude(self, value):
        self['AMPL'] = value

    @property
    def frequency(self):
        """."""
        return self['FREQ_ACT']

    @frequency.setter
    def frequency(self, value):
        self['FREQ'] = value

    @property
    def span(self):
        """."""
        return self['SPAN_ACT']

    @span.setter
    def span(self, value):
        self['SPAN'] = value

    @property
    def period(self):
        """."""
        return self['PERIOD_ACT']

    @period.setter
    def period(self, value):
        self['PERIOD'] = value

    @property
    def mask(self):
        """."""
        return self['MASK']

    @mask.setter
    def mask(self, value):
        self['MASK'] = _np.array(value)

    @property
    def mask_pattern(self):
        """."""
        return self['PATTERN']

    @mask_pattern.setter
    def mask_pattern(self, value):
        self['PATTERN'] = str(value)


class BunchClean(_ProptyDevice):
    """."""

    _properties = (
        'ENABLE', 'AMPL', 'TUNE', 'PATTERN', 'PERIOD', 'SPAN',
        )

    def __init__(self, devname):
        """."""
        devname = BunchbyBunch.process_device_name(devname)
        super().__init__(
            devname, propty_prefix='CLEAN_', properties=BunchClean._properties)

    @property
    def state(self):
        """."""
        return self['ENABLE']

    @state.setter
    def state(self, value):
        self['ENABLE'] = int(value)

    @property
    def amplitude(self):
        """."""
        return self['AMPL']

    @amplitude.setter
    def amplitude(self, value):
        self['AMPL'] = value

    @property
    def frequency(self):
        """."""
        return self['TUNE']

    @frequency.setter
    def frequency(self, value):
        self['TUNE'] = value

    @property
    def span(self):
        """."""
        return self['SPAN']

    @span.setter
    def span(self, value):
        self['SPAN'] = value

    @property
    def period(self):
        """."""
        return self['PERIOD']

    @period.setter
    def period(self, value):
        self['PERIOD'] = value

    @property
    def mask_pattern(self):
        """."""
        return self['PATTERN']

    @mask_pattern.setter
    def mask_pattern(self, value):
        self['PATTERN'] = str(value)


class PwrAmpL(_ProptyDevice):
    """."""

    _properties = ('FAULT', 'TEMP', 'FWDLOSS', 'REVLOSS', 'FWD', 'REV')

    DEF_TIMEOUT = 10  # [s]

    def __init__(self, devname, num=0):
        """."""
        devname = BunchbyBunch.process_device_name(devname)

        # call base class constructor
        super().__init__(
            devname, propty_prefix=f'MCLRAW_{num:d}_',
            properties=PwrAmpL._properties)

    @property
    def status(self):
        """."""
        return self['FAULT']

    @property
    def forward_loss(self):
        """."""
        return self['FWDLOSS']

    @forward_loss.setter
    def forward_loss(self, value):
        self['FWDLOSS'] = value

    @property
    def reverse_loss(self):
        """."""
        return self['REVLOSS']

    @reverse_loss.setter
    def reverse_loss(self, value):
        self['REVLOSS'] = value

    @property
    def forward(self):
        """."""
        return self['FWD']

    @property
    def reverse(self):
        """."""
        return self['REV']

    @property
    def temperature(self):
        """."""
        return self['TEMP']


class PwrAmpT(_Device):
    """."""

    DEF_TIMEOUT = 10  # [s]

    _properties = (
        'Rst-Cmd', 'Enbl-Sts', 'Enbl-Sel', 'GainAuto-Sts', 'GainAuto-Sel',
        'Gain-SP', 'Gain-RB', 'GainStep-SP', 'GainStep-RB',
        )

    def __init__(self, devname):
        """."""
        devname = BunchbyBunch.process_device_name(devname)
        devname = devname.substitute(dev='BbBAmp'+devname.idx, idx='')

        # call base class constructor
        super().__init__(devname, properties=PwrAmpT._properties)

    @property
    def enable(self):
        """."""
        return self['Enbl-Sts']

    @enable.setter
    def enable(self, value):
        self['Enbl-Sel'] = int(value)

    @property
    def auto_gain(self):
        """."""
        return self['GainAuto-Sts']

    @auto_gain.setter
    def auto_gain(self, value):
        self['GainAuto-Sel'] = int(value)

    @property
    def gain(self):
        """."""
        return self['Gain-RB']

    @gain.setter
    def gain(self, value):
        self['Gain-SP'] = value

    @property
    def gain_step(self):
        """."""
        return self['GainStep-RB']

    @gain_step.setter
    def gain_step(self, value):
        self['GainStep-SP'] = value

    def cmd_reset(self, timeout=DEF_TIMEOUT):
        """."""
        self['Rst-Cmd'] = 1
        if not self._wait('Rst-Cmd', 1, timeout/2):
            return False
        _time.sleep(0.2)
        self['Rst-Cmd'] = 0
        return self._wait('Rst-Cmd', 0, timeout/2)



"""PosAng Device."""

from .device import Device as _Device


class PosAng(_Device):
    """Device to control position and angle correction IOC."""

    class DEVICES:
        """Devices names."""

        TB = 'TB-Glob:AP-PosAng'
        TS = 'TS-Glob:AP-PosAng'
        ALL = (TB, TS)

    _properties = (
        'DeltaPosX-SP', 'DeltaPosX-RB',
        'DeltaAngX-SP', 'DeltaAngX-RB',
        'DeltaPosY-SP', 'DeltaPosY-RB',
        'DeltaAngY-SP', 'DeltaAngY-RB',
        'SetNewRefKick-Cmd', 'NeedRefUpdate-Mon',
        'ConfigName-SP', 'ConfigName-RB',
        'RespMatX-Mon', 'RespMatY-Mon',
    )

    def __init__(self, devname):
        """Init."""
        # check if device exists
        if devname not in PosAng.DEVICES.ALL:
            raise NotImplementedError(devname)

        # call base class constructor
        super().__init__(devname, properties=PosAng._properties)

    @property
    def delta_posx(self):
        """Delta PosX."""
        return self['DeltaPosX-RB']

    @delta_posx.setter
    def delta_posx(self, value):
        self['DeltaPosX-SP'] = value

    @property
    def delta_angx(self):
        """Delta AngX."""
        return self['DeltaAngX-RB']

    @delta_angx.setter
    def delta_angx(self, value):
        self['DeltaAngX-SP'] = value

    @property
    def delta_posy(self):
        """Delta PosY."""
        return self['DeltaPosY-RB']

    @delta_posy.setter
    def delta_posy(self, value):
        self['DeltaPosY-SP'] = value

    @property
    def delta_angy(self):
        """Delta AngY."""
        return self['DeltaAngY-RB']

    @delta_angy.setter
    def delta_angy(self, value):
        self['DeltaAngY-SP'] = value

    @property
    def config_name(self):
        """PosAng configuration name."""
        return self['ConfigName-RB']

    @config_name.setter
    def config_name(self, value):
        self['ConfigName-SP'] = value

    @property
    def respmat_x(self):
        """Response matrix for PosAngX."""
        return self['RespMatX-Mon'].reshape(2, -1)

    @property
    def respmat_y(self):
        """Response matrix for PosAngY."""
        return self['RespMatY-Mon'].reshape(2, -1)

    @property
    def need_ref_update(self):
        """Need Reference Update."""
        return self['NeedRefUpdate-Mon']

    def cmd_update_reference(self):
        """Update PosAng reference kicks."""
        self['SetNewRefKick-Cmd'] = 1
        return True



"""Power Supply Module."""

# import random as _random
import time as _time
import math as _math
import numpy as _np


from ...bsmp import SerialError as _SerialError


def _psupply_update_connected(func):
    # This functions decorates PSDevState methods to keep track of
    # connected status.
    def new_update_func(obj, *args, **kwargs):
        try:
            response = func(obj, *args, **kwargs)
            setattr(obj, '_connected', True)
            return response
        except _SerialError:
            setattr(obj, '_connected', False)
            raise
    return new_update_func


class PSDevState:
    """Power supply device state.

        Objects of this class are used to store power supply device
    states.
    """

    DEFAULT_UPDATE_INTERVAL_WFM = 2.0  # [s]
    DEFAULT_UPDATE_INTERVAL_VARIABLES = 0.1  # [s]
    DEFAULT_UPDATE_INTERVAL_PARAMETERS = 2.0  # [s]

    def __init__(self, psbsmp):
        """Init."""
        self._psbsmp = psbsmp
        self._connected = None
        self._groups = PSDevState._init_groups()
        self._variables = PSDevState._init_variables()
        self._curves = PSDevState._init_curves()
        self._parameters = self._init_parameters()
        self._wfmref_rb = None
        self._wfmref = None
        self._scope = None
        self._timestamp_update = None
        self._timestamp_update_group = None
        self._timestamp_update_variables = None
        self._timestamp_update_wfm = None
        self._timestamp_update_parameters = None

    @property
    def psbsmp(self):
        """Return PSBSMP communication objetc."""
        return self._psbsmp

    @property
    def connected(self):
        """Return connection state."""
        return self._connected

    @property
    def wfmref_rb(self):
        """Return wfmref_rb (Wfm-RB)."""
        return self._wfmref_rb

    @wfmref_rb.setter
    def wfmref_rb(self, value):
        """Set wfmref_rb.

        This is invoked for the object to store wfmref that has been read
        from power supply right after writting a new wfmref curve.
        """
        self._wfmref_rb = _np.array(value)

    @property
    def wfmref(self):
        """Return wfmref (WfmRef-Mon)."""
        return self._wfmref

    @property
    def wfmref_index(self):
        """Return current index into DSP selected curve."""
        curve_id = \
            self._variables[self._psbsmp.CONST.V_WFMREF_SELECTED]
        if curve_id == 0:
            beg = self._variables[
                self._psbsmp.CONST.V_WFMREF0_START]
            end = self._variables[
                self._psbsmp.CONST.V_WFMREF0_END]
        else:
            beg = self._variables[
                self._psbsmp.CONST.V_WFMREF1_START]
            end = self._variables[
                self._psbsmp.CONST.V_WFMREF1_END]
        index = self._psbsmp.curve_index_calc(beg, end)
        return index

    @property
    def scope(self):
        """Return scope (Wfm-Mon)."""
        return self._scope

    @property
    def variables(self):
        """."""
        return self._variables

    @property
    def parameters(self):
        """."""
        return self._parameters

    @property
    def timestamp_update(self):
        """Return timestamp of last power supply update."""
        tstamps = (
            self._timestamp_update,
            self._timestamp_update_group,
            self._timestamp_update_variables,
            self._timestamp_update_wfm)
        maxtstamp = max(val if val is not None else 0 for val in tstamps)
        return maxtstamp

    def get_variable(self, var_id):
        """."""
        return self._variables[var_id]

    def get_parameter(self, eid):
        """."""
        return self._parameters[eid]

    def update(self, interval=0.0):
        """Update all power supply entities.

        Method does not update groups of variables.
        """
        now = _time.time()
        tstamp = self._timestamp_update
        if tstamp is None or (now - tstamp) > interval:
            connected = True
            # variables
            if not self.update_variables(interval=0.0):
                return False
            connected &= self._connected
            # wfmref
            if not self.update_wfm(interval=0.0):
                return False
            connected &= self._connected
            # parameters
            if not self.update_parameters(interval=0.0):
                return False
            connected &= self._connected
            # update connected state
            self._connected = connected
            if connected:
                self._timestamp_update = now
        return True

    @_psupply_update_connected
    def update_groups(self, interval=0.0):
        """."""
        now = _time.time()
        tstamp = self._timestamp_update_group
        if tstamp is None or (now - tstamp) >= interval:
            ack, groupsinfo = self._psbsmp.query_list_of_group_of_variables()
            if ack == self.psbsmp.CONST_BSMP.ACK_OK:
                for i in range(len(groupsinfo)):
                    ack, var_ids = \
                        self._psbsmp.query_group_of_variables(group_id=i)
                    if ack == self.psbsmp.CONST_BSMP.ACK_OK:
                        self._groups[i] = var_ids
                    else:
                        return False
            else:
                return False
            self._timestamp_update_group = now
        return True

    @_psupply_update_connected
    def update_variables(self, interval=None):
        """Update all variables."""
        if interval is None:
            interval = PSDevState.DEFAULT_UPDATE_INTERVAL_VARIABLES
        now = _time.time()
        tstamp = self._timestamp_update_variables
        if tstamp is None or (now - tstamp) >= interval:
            group_id = self._psbsmp.CONST.G_ALL
            ack, values = self._psbsmp.read_group_of_variables(
                group_id=group_id)
            if ack == self.psbsmp.CONST_BSMP.ACK_OK:
                # update _variables with data read from device
                if group_id not in self._groups:
                    return False
                var_ids = self._groups[group_id]
                for var_id, value in zip(var_ids, values):
                    self._variables[var_id] = value
                # NOTE: test line for benchmarking
                # self._variables[27] += 0.001 * (_random.random() - 0.5)
                # update timestamp
                self._timestamp_update_variables = now
            else:
                return False
        return True

    @_psupply_update_connected
    def update_wfm(self, interval=None):
        """Update wfmref."""
        if self._psbsmp.IS_DCLINK:
            return True
        if interval is None:
            interval = PSDevState.DEFAULT_UPDATE_INTERVAL_WFM
        now = _time.time()
        tstamp = self._timestamp_update_wfm
        if tstamp is None or (now - tstamp) >= interval:
            # update wfmref
            wfmref = self._psbsmp.wfmref_read()
            if wfmref is None:
                return False
            self._wfmref = wfmref
            # update scope
            scope = self._psbsmp.scope_read()
            if scope is not None:
                self._scope = scope
                # NOTE: a successful readout means scope is not
                # currently being filled by DSP. Therefore a scope-fill
                # enable command is in order!
                self._psbsmp.scope_enable()
            # update timestamp (even it could not read scope)
            self._timestamp_update_wfm = now
        return True

    @_psupply_update_connected
    def update_parameters(self, interval=None):
        """Update power supply parameters."""
        if interval is None:
            interval = PSDevState.DEFAULT_UPDATE_INTERVAL_PARAMETERS
        now = _time.time()
        tstamp = self._timestamp_update_parameters
        if tstamp is None or (now - tstamp) >= interval:
            parameters = self._psbsmp.entities.parameters
            conn_status_ok = True
            for eid in parameters.eids:
                parameter = parameters[eid]
                counter = parameter['count']
                if eid != self._psbsmp.CONST.P_PS_NAME and counter > 1:
                    value = _np.zeros(counter)
                    for idx in range(counter):
                        value[idx] = self._psbsmp.parameter_read(eid, idx)
                        if value[idx] is None:
                            conn_status_ok = False
                        if _math.isnan(value[idx]):
                            value[idx] = 0
                else:
                    value = self._psbsmp.parameter_read(eid)
                    if value is None:
                        conn_status_ok = False
                self._parameters[eid] = value
                # print(eid, value)
            if conn_status_ok:
                self._timestamp_update_parameters = now
                return True
            return False
        return True

    def reset_variables_groups(self, groups):
        """."""
        # reset groups of variables for all bsmp devices
        ack, _ = self._psbsmp.reset_groups_of_variables(groups[3:])
        if ack != self.psbsmp.CONST_BSMP.ACK_OK:
            ValueError('Could not reset groups of variables!')
        return True

    # --- private methods ---

    @staticmethod
    def _init_groups():
        # NOTE: template to be expanded, if necessary.
        return dict()

    @staticmethod
    def _init_variables():
        # NOTE: template to be expanded, if necessary.
        return dict()

    @staticmethod
    def _init_curves():
        # NOTE: template to be expanded, if necessary.
        return dict()

    def _init_parameters(self):
        """Init power supply parameters."""
        parameters = self._psbsmp.entities.parameters
        parameters_values = dict()
        for eid in parameters.eids:
            if eid == 0:
                parameters_values[eid] = ''
            else:
                parameter = parameters[eid]
                counter = parameter['count']
                value = _np.zeros(counter) * float('NaN')
                parameters_values[eid] = value
        return parameters_values



"""Insertion Device Feedforward Correction Classes."""

import math as _math
import numpy as _np

from ..search import IDSearch as _IDSearch
from ..search import MASearch as _MASearch


class APUFFWDCalc:
    """."""

    def __init__(self, idname):
        """."""
        # get correctors names
        self._psnames_orb = _IDSearch.conv_idname_2_orbitcorr(idname)
        # NOTE: assumes same number of CHs and CVs
        self._nr_chs = len(self._psnames_orb) // 2
        self._nr_cvs = self._nr_chs

        # get corr spos
        self._orbcorr_spos = self._get_corr_spos()

        # get orbit fftable from idname:
        self._orbitffwd = _IDSearch.conv_idname_2_orbitffwd(idname)

    @property
    def nr_chs(self):
        """Return number of orbit CH correctors."""
        return self._nr_chs

    @property
    def nr_cvs(self):
        """Return number of orbit CV correctors."""
        return self._nr_cvs

    @property
    def orbitcorr_psnames(self):
        """Return orbit corrector names."""
        return self._psnames_orb

    def conv_phase_2_orbcorr_currents(self, phase):
        """Return orbit correctors currents for a given ID phase."""
        ffwd = self._orbitffwd.interp_curr2mult(phase)
        chs = [ffwd['normal'][i] for i in range(self.nr_chs)]
        cvs = [ffwd['skew'][i] for i in range(self.nr_cvs)]
        currents = _np.array(chs + cvs)
        return currents

    def conv_posang2kick(
            self, posx=0, angx=0, posy=0, angy=0):
        """Return orbit correctors currents for bumps and angles."""
        # 3.34 µs ± 49.1 ns per loop
        spos = self._orbcorr_spos
        len1 = spos[1] - spos[0]
        len2 = 0.5*(spos[2] - spos[1])
        kickx = APUFFWDCalc._calc_kicks(len1, len2, posx, angx)
        kicky = APUFFWDCalc._calc_kicks(len1, len2, posy, angy)
        return _np.asarray(kickx + kicky)

    # --- private methods ---

    @staticmethod
    def _calc_kicks(len1, len2, pos, ang):
        """Geometry (a courtesy of F. de Sá).

        >----------------> ebeam direction >---------------->
        C1|C1      C2|C2                     C3|C3      C4|C4
          |---len1---|----len2----|----len2----|---len1---|

        """
        # ang bump
        theta = _math.atan(len2/len1 * _math.tan(ang/1e6)) * 1e6
        kicks = [-theta, theta + ang, -theta - ang, theta]
        # pos bump
        theta = _math.atan(pos / 1e6 / len1) * 1e6
        kicks[0] += theta
        kicks[1] -= theta
        kicks[2] -= theta
        kicks[3] += theta
        return kicks

    def _get_corr_spos(self):
        manames = [psname.replace(':PS-', ':MA-') for
                   psname in self._psnames_orb]
        spos = _MASearch.get_mapositions(names=manames)
        return spos



"""BO ramp configuration.

Values in _template_dict are arbitrary. They are used just to compare with
corresponding values when a new configuration is tried to be inserted in the
servconf database.
"""
from copy import deepcopy as _dcopy
from siriuspy.pwrsupply.csdev import \
    DEF_WFMSIZE_FBP as _DEF_WFMSIZE_FBP, \
    DEF_WFMSIZE_OTHERS as _DEF_WFMSIZE_OTHERS

from siriuspy.ramp import util as _util


def get_dict():
    """Return a dict with ramp settings."""
    module_name = __name__.split('.')[-1]
    _dict = {
        'config_type_name': module_name,
        'value': _dcopy(_template_dict)
    }

    return _dict


_ps_ramp = {
    # number of points in power supply waveforms
    'wfm_nrpoints_fams': _DEF_WFMSIZE_OTHERS,
    'wfm_nrpoints_corrs': _DEF_WFMSIZE_FBP,
    # ramp total duration [ms]
    'duration': _util.DEFAULT_PS_RAMP_DURATION,
    # ramp time parameters [ms]
    'rampup1_start_time': _util.DEFAULT_PS_RAMP_RAMPUP1_START_TIME,
    'rampup2_start_time': _util.DEFAULT_PS_RAMP_RAMPUP2_START_TIME,
    'rampdown_start_time': _util.DEFAULT_PS_RAMP_RAMPDOWN_START_TIME,
    'rampdown_stop_time': _util.DEFAULT_PS_RAMP_RAMPDOWN_STOP_TIME,
    'rampup_smooth_intvl': _util.DEFAULT_PS_RAMP_RAMPUP_SMOOTH_INTVL,
    'rampdown_smooth_intvl': _util.DEFAULT_PS_RAMP_RAMPDOWN_SMOOTH_INTVL,
    # ramp energy parameters [GeV]
    'start_energy': _util.DEFAULT_PS_RAMP_START_ENERGY,
    'rampup1_start_energy': _util.DEFAULT_PS_RAMP_RAMPUP1_START_ENERGY,
    'rampup2_start_energy': _util.DEFAULT_PS_RAMP_RAMPUP2_START_ENERGY,
    'rampdown_start_energy': _util.DEFAULT_PS_RAMP_RAMPDOWN_START_ENERGY,
    'rampdown_stop_energy': _util.DEFAULT_PS_RAMP_RAMPDOWN_STOP_ENERGY,
    'rampup_smooth_energy': _util.DEFAULT_PS_RAMP_RAMPUP_SMOOTH_ENERGY,
    'rampdown_smooth_energy': _util.DEFAULT_PS_RAMP_RAMPDOWN_SMOOTH_ENERGY,
}

# time [ms]            normalized configuration name
_ps_normalized_configs = {}

_rf_ramp = {
    # ramp intervals durations [ms]
    'bottom_duration':      _util.DEFAULT_RF_RAMP_BOTTOM_DURATION,
    'rampup_duration':      _util.DEFAULT_RF_RAMP_RAMPUP_DURATION,
    'top_duration':         _util.DEFAULT_RF_RAMP_TOP_DURATION,
    'rampdown_duration':    _util.DEFAULT_RF_RAMP_RAMPDOWN_DURATION,
    # ramp voltage parameters [kV]
    'bottom_voltage':       _util.DEFAULT_RF_RAMP_BOTTOM_VOLTAGE,
    'top_voltage':          _util.DEFAULT_RF_RAMP_TOP_VOLTAGE,
    # ramp phase parameters [°]
    'bottom_phase':         _util.DEFAULT_RF_RAMP_BOTTOM_PHASE,
    'top_phase':            _util.DEFAULT_RF_RAMP_TOP_PHASE,
}

_ti_params = {
    'injection_time':       _util.DEFAULT_TI_PARAMS_INJECTION_TIME,  # [ms]
    'ejection_time':        _util.DEFAULT_TI_PARAMS_EJECTION_TIME,  # [ms]
    'rf_ramp_delay':        _util.DEFAULT_TI_PARAMS_RF_RAMP_DELAY,  # [us]
    'ps_ramp_delay':        _util.DEFAULT_TI_PARAMS_PS_RAMP_DELAY,  # [us]
}


_template_dict = {
    'ps_ramp': _ps_ramp,
    'ps_normalized_configs*': _ps_normalized_configs,
    'rf_ramp': _rf_ramp,
    'ti_params': _ti_params,
}



"""Define PVs, constants and properties of OpticsCorr SoftIOCs."""
from .. import csdev as _csdev


# --- Enumeration Types ---

class ETypes(_csdev.ETypes):
    """Local enumerate types."""

    PROP_ADD = ('Proportional', 'Additional')
    INDIV_2KNOBS = ('Individual', 'TwoKnobs')
    MEAS_CMD = ('Reset', 'Start', 'Stop')
    MEAS_MON = ('Idle', 'Measuring', 'Completed', 'Aborted')


_et = ETypes  # syntactic sugar


# --- Const class ---

class Const(_csdev.Const):
    """Const class defining OpticsCorr constants and Enum types."""

    CorrMeth = _csdev.Const.register('CorrMeth', _et.PROP_ADD)
    CorrGroup = _csdev.Const.register('CorrGroup', _et.INDIV_2KNOBS)
    SyncCorr = _csdev.Const.register('SyncCorr', _et.OFF_ON)
    MeasCmd = _csdev.Const.register('MeasCmd', _et.MEAS_CMD)
    MeasMon = _csdev.Const.register('MeasMon', _et.MEAS_MON)

    BO_SFAMS_CHROMCORR = ('SF', 'SD')
    BO_SFAMS_NELM = (25, 10)

    SI_SFAMS_CHROMCORR = ('SFA1', 'SFA2', 'SDA1', 'SDA2', 'SDA3',
                          'SFB1', 'SFB2', 'SDB1', 'SDB2', 'SDB3',
                          'SFP1', 'SFP2', 'SDP1', 'SDP2', 'SDP3')
    SI_SFAMS_NELM = (10, 10, 10, 10, 10,
                     20, 20, 20, 20, 20,
                     10, 10, 10, 10, 10)

    BO_QFAMS_TUNECORR = ('QF', 'QD')
    BO_QFAMS_NELM = (50, 25)

    SI_QFAMS_TUNECORR = ('QFA', 'QFB', 'QFP',
                         'QDA', 'QDB1', 'QDB2', 'QDP1', 'QDP2')
    SI_QFAMS_NELM = (10, 20, 10,
                     10, 20, 20, 10, 10)

    STATUS_LABELS = ('PS Connection', 'PS PwrState', 'PS OpMode',
                     'PS CtrlMode', 'Timing Config')


_ct = Const  # syntactic sugar


# --- Databases ---


def get_chrom_database(acc):
    """Return OpticsCorr-Chrom Soft IOC database."""
    if acc == 'BO':
        sfams = _ct.BO_SFAMS_CHROMCORR
    elif acc == 'SI':
        sfams = _ct.SI_SFAMS_CHROMCORR

    corrmat_size = len(sfams)*2

    pvs_database = {
        'Version-Cte': {'type': 'string', 'value': 'UNDEF'},

        'Log-Mon': {'type': 'string', 'value': 'Starting...'},

        'ChromX-SP': {'type': 'float', 'value': 0, 'prec': 6,
                      'hilim': 10, 'lolim': -10, 'high': 10,
                      'low': -10, 'hihi': 10, 'lolo': -10},
        'ChromX-RB': {'type': 'float', 'value': 0, 'prec': 6},
        'ChromY-SP': {'type': 'float', 'value': 0, 'prec': 6,
                      'hilim': 10, 'lolim': -10, 'high': 10,
                      'low': -10, 'hihi': 10, 'lolo': -10},
        'ChromY-RB': {'type': 'float', 'value': 0, 'prec': 6},

        'DeltaChromX-SP': {'type': 'float', 'value': 0, 'prec': 6,
                           'hilim': 10, 'lolim': -10, 'high': 10,
                           'low': -10, 'hihi': 10, 'lolo': -10},
        'DeltaChromX-RB': {'type': 'float', 'value': 0, 'prec': 6},
        'DeltaChromY-SP': {'type': 'float', 'value': 0, 'prec': 6,
                           'hilim': 10, 'lolim': -10, 'high': 10,
                           'low': -10, 'hihi': 10, 'lolo': -10},
        'DeltaChromY-RB': {'type': 'float', 'value': 0, 'prec': 6},

        'ChromX-Mon': {'type': 'float', 'value': 0, 'prec': 6},
        'ChromY-Mon': {'type': 'float', 'value': 0, 'prec': 6},
        'CalcChromX-Mon': {'type': 'float', 'value': 0, 'prec': 6},
        'CalcChromY-Mon': {'type': 'float', 'value': 0, 'prec': 6},

        'ApplyDelta-Cmd': {'type': 'int', 'value': 0},

        'ConfigName-SP': {'type': 'string', 'value': ''},
        'ConfigName-RB': {'type': 'string', 'value': ''},
        'RespMat-Mon': {'type': 'float', 'count': corrmat_size,
                        'value': corrmat_size*[0], 'prec': 6, 'unit':
                        'Chrom x SFams (Matrix of add method)'},
        'NominalChrom-Mon': {'type': 'float', 'count': 2, 'value': 2*[0],
                             'prec': 6},
        'NominalSL-Mon': {'type': 'float', 'count': len(sfams),
                          'value': len(sfams)*[0], 'prec': 6},

        'ConfigPS-Cmd': {'type': 'int', 'value': 0},

        'Status-Mon': {'type': 'int', 'value': 0b11111},
        'StatusLabels-Cte': {'type': 'char', 'count': 1000,
                             'value': '\n'.join(_ct.STATUS_LABELS)},
    }

    for fam in sfams:
        pvs_database['SL' + fam + '-Mon'] = {
            'type': 'float', 'value': 0, 'prec': 4, 'unit': '1/m^2',
            'lolim': 0, 'hilim': 0, 'low': 0, 'high': 0, 'lolo': 0, 'hihi': 0}

    if acc == 'SI':
        pvs_database['CorrMeth-Sel'] = {
            'type': 'enum', 'enums': _et.PROP_ADD,
            'value': _ct.CorrMeth.Proportional}
        pvs_database['CorrMeth-Sts'] = {
            'type': 'enum', 'enums': _et.PROP_ADD,
            'value': _ct.CorrMeth.Proportional}
        pvs_database['CorrGroup-Sel'] = {
            'type': 'enum', 'enums': _et.INDIV_2KNOBS,
            'value': _ct.CorrGroup.TwoKnobs}
        pvs_database['CorrGroup-Sts'] = {
            'type': 'enum', 'enums': _et.INDIV_2KNOBS,
            'value': _ct.CorrGroup.TwoKnobs}

        # pvs_database['SyncCorr-Sel'] = {
        #     'type': 'enum', 'enums': _et.OFF_ON, 'value': _ct.SyncCorr.Off}
        # pvs_database['SyncCorr-Sts'] = {
        #     'type': 'enum', 'enums': _et.OFF_ON, 'value': _ct.SyncCorr.Off}
        # pvs_database['ConfigTiming-Cmd'] = {'type': 'int', 'value': 0}

        pvs_database['MeasChromDeltaFreqRF-SP'] = {
            'type': 'float', 'value': 200.0, 'unit': 'Hz',
            'prec': 4, 'lolim': 0.1, 'hilim': 1000.0}
        pvs_database['MeasChromDeltaFreqRF-RB'] = {
            'type': 'float', 'value': 200.0, 'unit': 'Hz',
            'prec': 4, 'lolim': 0.1, 'hilim': 1000.0}
        pvs_database['MeasChromWaitTune-SP'] = {
            'type': 'float', 'value': 5, 'unit': 's',
            'prec': 3, 'lolim': 0.005, 'hilim': 100}
        pvs_database['MeasChromWaitTune-RB'] = {
            'type': 'float', 'value': 5, 'unit': 's',
            'prec': 3, 'lolim': 0.005, 'hilim': 100}
        pvs_database['MeasChromNrSteps-SP'] = {
            'type': 'int', 'value': 8, 'unit': 'nr pts',
            'prec': 3, 'lolim': 2, 'hilim': 10}
        pvs_database['MeasChromNrSteps-RB'] = {
            'type': 'int', 'value': 8, 'unit': 'nr pts',
            'prec': 3, 'lolim': 2, 'hilim': 10}
        pvs_database['MeasChrom-Cmd'] = {
            'type': 'enum', 'enums': _et.MEAS_CMD, 'value': _ct.MeasCmd.Reset}
        pvs_database['MeasChromStatus-Mon'] = {
            'type': 'enum', 'enums': _et.MEAS_MON, 'value': _ct.MeasMon.Idle}
        pvs_database['MeasChromX-Mon'] = {
            'type': 'float', 'value': 0, 'prec': 6}
        pvs_database['MeasChromY-Mon'] = {
            'type': 'float', 'value': 0, 'prec': 6}

        pvs_database['MeasConfigDeltaSLFamSF-SP'] = {
            'type': 'float', 'value': 10.000, 'unit': '1/m2',
            'prec': 3, 'lolim': 0.100, 'hilim': 50.000}
        pvs_database['MeasConfigDeltaSLFamSF-RB'] = {
            'type': 'float', 'value': 10.000, 'unit': '1/m2',
            'prec': 3, 'lolim': 0.100, 'hilim': 50.000}
        pvs_database['MeasConfigDeltaSLFamSD-SP'] = {
            'type': 'float', 'value': 10.000, 'unit': '1/m2',
            'prec': 3, 'lolim': 0.100, 'hilim': 50.000}
        pvs_database['MeasConfigDeltaSLFamSD-RB'] = {
            'type': 'float', 'value': 10.000, 'unit': '1/m2',
            'prec': 3, 'lolim': 0.100, 'hilim': 50.000}
        pvs_database['MeasConfigWait-SP'] = {
            'type': 'float', 'value': 1, 'unit': 's',
            'prec': 3, 'lolim': 0.005, 'hilim': 100}
        pvs_database['MeasConfigWait-RB'] = {
            'type': 'float', 'value': 1, 'unit': 's',
            'prec': 3, 'lolim': 0.005, 'hilim': 100}
        pvs_database['MeasConfigName-SP'] = {
            'type': 'string', 'value': 'UNDEF'}
        pvs_database['MeasConfigName-RB'] = {
            'type': 'string', 'value': 'UNDEF'}
        pvs_database['MeasConfigSave-Cmd'] = {
            'type': 'int', 'value': 0}
        pvs_database['MeasConfig-Cmd'] = {
            'type': 'enum', 'enums': _et.MEAS_CMD, 'value': _ct.MeasCmd.Reset}
        pvs_database['MeasConfigStatus-Mon'] = {
            'type': 'enum', 'enums': _et.MEAS_MON, 'value': _ct.MeasMon.Idle}

    pvs_database = _csdev.add_pvslist_cte(pvs_database)
    return pvs_database


def get_tune_database(acc):
    """Return OpticsCorr-Tune Soft IOC database."""
    if acc == 'BO':
        qfams = _ct.BO_QFAMS_TUNECORR
    elif acc == 'SI':
        qfams = _ct.SI_QFAMS_TUNECORR

    corrmat_size = len(qfams)*2

    pvs_database = {
        'Version-Cte': {'type': 'string', 'value': 'UNDEF'},

        'Log-Mon': {'type': 'string', 'value': 'Starting...'},

        'DeltaTuneX-SP': {'type': 'float', 'value': 0, 'prec': 6,
                          'hilim': 5, 'lolim': -5, 'high': 5, 'low': -5,
                          'hihi': 5, 'lolo': -5},
        'DeltaTuneX-RB': {'type': 'float', 'value': 0, 'prec': 6,
                          'hilim': 5, 'lolim': -5, 'high': 5, 'low': -5,
                          'hihi': 5, 'lolo': -5},
        'DeltaTuneX-Mon': {'type': 'float', 'value': 0, 'prec': 6,
                           'hilim': 5, 'lolim': -5, 'high': 5, 'low': -5,
                           'hihi': 5, 'lolo': -5},
        'DeltaTuneY-SP': {'type': 'float', 'value': 0, 'prec': 6,
                          'hilim': 5, 'lolim': -5, 'high': 5, 'low': -5,
                          'hihi': 5, 'lolo': -5},
        'DeltaTuneY-RB': {'type': 'float', 'value': 0, 'prec': 6,
                          'hilim': 5, 'lolim': -5, 'high': 5, 'low': -5,
                          'hihi': 5, 'lolo': -5},
        'DeltaTuneY-Mon': {'type': 'float', 'value': 0, 'prec': 6,
                           'hilim': 5, 'lolim': -5, 'high': 5, 'low': -5,
                           'hihi': 5, 'lolo': -5},

        'ApplyDelta-Cmd': {'type': 'int', 'value': 0},

        'ConfigName-SP': {'type': 'string', 'value': ''},
        'ConfigName-RB': {'type': 'string', 'value': ''},
        'RespMat-Mon': {'type': 'float', 'count': corrmat_size,
                        'value': corrmat_size*[0], 'prec': 6, 'unit':
                        'Tune x QFams (Nominal Response Matrix)'},
        'NominalKL-Mon': {'type': 'float', 'count': len(qfams),
                          'value': len(qfams)*[0], 'prec': 6},

        'ConfigPS-Cmd': {'type': 'int', 'value': 0},

        'SetNewRefKL-Cmd': {'type': 'int', 'value': 0},

        'Status-Mon': {'type': 'int', 'value': 0b11111},
        'StatusLabels-Cte': {'type': 'char', 'count': 1000,
                             'value': '\n'.join(_ct.STATUS_LABELS)},
    }

    for fam in qfams:
        pvs_database['RefKL' + fam + '-Mon'] = {
            'type': 'float', 'value': 0, 'prec': 6, 'unit': '1/m'}

        pvs_database['DeltaKL' + fam + '-Mon'] = {
            'type': 'float', 'value': 0, 'prec': 6, 'unit': '1/m',
            'lolim': 0, 'hilim': 0, 'low': 0, 'high': 0, 'lolo': 0, 'hihi': 0}

    if acc == 'SI':
        pvs_database['CorrMeth-Sel'] = {
            'type': 'enum', 'enums': _et.PROP_ADD,
            'value': _ct.CorrMeth.Proportional}
        pvs_database['CorrMeth-Sts'] = {
            'type': 'enum', 'enums': _et.PROP_ADD,
            'value': _ct.CorrMeth.Proportional}
        pvs_database['CorrGroup-Sel'] = {
            'type': 'enum', 'enums': _et.INDIV_2KNOBS,
            'value': _ct.CorrGroup.TwoKnobs}
        pvs_database['CorrGroup-Sts'] = {
            'type': 'enum', 'enums': _et.INDIV_2KNOBS,
            'value': _ct.CorrGroup.TwoKnobs}

        pvs_database['SyncCorr-Sel'] = {
            'type': 'enum', 'enums': _et.OFF_ON, 'value': _ct.SyncCorr.Off}
        pvs_database['SyncCorr-Sts'] = {
            'type': 'enum', 'enums': _et.OFF_ON, 'value': _ct.SyncCorr.Off}
        pvs_database['ConfigTiming-Cmd'] = {'type': 'int', 'value': 0}

        pvs_database['MeasConfigDeltaKLFamQF-SP'] = {
            'type': 'float', 'value': 0.020, 'unit': '1/m',
            'prec': 3, 'lolim': 0.001, 'hilim': 0.100}
        pvs_database['MeasConfigDeltaKLFamQF-RB'] = {
            'type': 'float', 'value': 0.020, 'unit': '1/m',
            'prec': 3, 'lolim': 0.001, 'hilim': 0.100}
        pvs_database['MeasConfigDeltaKLFamQD-SP'] = {
            'type': 'float', 'value': 0.020, 'unit': '1/m',
            'prec': 3, 'lolim': 0.001, 'hilim': 0.100}
        pvs_database['MeasConfigDeltaKLFamQD-RB'] = {
            'type': 'float', 'value': 0.020, 'unit': '1/m',
            'prec': 3, 'lolim': 0.001, 'hilim': 0.100}
        pvs_database['MeasConfigWait-SP'] = {
            'type': 'float', 'value': 1, 'unit': 's',
            'prec': 3, 'lolim': 0.005, 'hilim': 100}
        pvs_database['MeasConfigWait-RB'] = {
            'type': 'float', 'value': 1, 'unit': 's',
            'prec': 3, 'lolim': 0.005, 'hilim': 100}
        pvs_database['MeasConfigName-SP'] = {
            'type': 'string', 'value': 'UNDEF'}
        pvs_database['MeasConfigName-RB'] = {
            'type': 'string', 'value': 'UNDEF'}
        pvs_database['MeasConfigSave-Cmd'] = {
            'type': 'int', 'value': 0}
        pvs_database['MeasConfig-Cmd'] = {
            'type': 'enum', 'enums': _et.MEAS_CMD, 'value': _ct.MeasCmd.Reset}
        pvs_database['MeasConfigStatus-Mon'] = {
            'type': 'enum', 'enums': _et.MEAS_MON, 'value': _ct.MeasMon.Idle}

    pvs_database = _csdev.add_pvslist_cte(pvs_database)
    return pvs_database



"""Module with connector classes.

This module implements connector classes responsible for communications with
pwrsupply soft IOcs, ConfigDB service and orbit IOCs.
"""

# TODO: Implementations in this module make pylint very confused.
# Is there a code efficiency reason to initialize class attributes
# outside private methods?

import math as _math
import numpy as _np


from .. import envars as _envars
from .. import csdev as _csdev

from ..epics import EpicsProperty as _EpicsProperty, \
    EpicsPropertiesList as _EpicsPropsList
from ..namesys import SiriusPVName as _PVName
from ..pwrsupply.csdev import Const as _PSConst
from ..timesys.csdev import Const as _TIConst, \
    get_hl_trigger_database as _get_trig_db
from ..sofb.csdev import SOFBRings as _SOFBRings
from ..search import LLTimeSearch as _LLTimeSearch, \
    PSSearch as _PSSearch

from . import util as _rutil


_PREFIX = _envars.VACA_PREFIX
_TIMEOUT_DFLT = 8
_TIMEOUT_PWRSTATE_ON = 2
_TIMEOUT_PWRSTATE_OFF = 1
_TIMEOUT_OPMODE_CHANGE = 6


class ConnTI(_EpicsPropsList):
    """Timing connector class."""

    class Const(_csdev.Const):
        """Properties names."""

        BO_HarmNum = 828

        # EVG PVs
        EVG = _LLTimeSearch.get_evg_name()
        EVG_DevEnbl = EVG + ':DevEnbl-Sel'
        EVG_ContinuousEvt = EVG + ':ContinuousEvt-Sel'
        EVG_InjectionEvt = EVG + ':InjectionEvt-Sel'
        EVG_UpdateEvt = EVG + ':UpdateEvt-Cmd'

        # Linac Egun mode properties
        LinacEgun_SglBun_State = 'LI-01:EG-PulsePS:singleselstatus'
        LinacEgun_MultBun_State = 'LI-01:EG-PulsePS:multiselstatus'

        # Interlock PV
        Intlk = 'LA-RFH01RACK2:TI-EVR:IntlkStatus-Mon'

        # RFFreq
        RFFreq = 'RF-Gen:GeneralFreq-RB'

    # Add events properties to Const
    _events = {
        'EvtLinac': Const.EVG + ':Linac',
        'EvtInjBO': Const.EVG + ':InjBO',
        'EvtInjSI': Const.EVG + ':InjSI',
        'EvtRmpBO': Const.EVG + ':RmpBO',
        'EvtStudy': Const.EVG + ':Study'}

    evt_propties = ('Mode-Sel', 'DelayType-Sel', 'Delay-SP')
    for attr, evt_name in _events.items():
        setattr(Const, attr, evt_name)
        for p in evt_propties:
            p = _PVName(p)
            new_attr = attr+'_'+p.propty_name
            setattr(Const, new_attr, evt_name+p)

    # Add trigger properties to Const
    _triggers = {
        'TrgMags': 'BO-Glob:TI-Mags-Fams',
        'TrgCorrs': 'BO-Glob:TI-Mags-Corrs',
        'TrgLLRFRmp': 'BO-Glob:TI-LLRF-Rmp',
        'TrgEGunSglBun': 'LI-01:TI-EGun-SglBun',
        'TrgEGunMultBun': 'LI-01:TI-EGun-MultBun',
        'TrgEjeKckr': 'BO-48D:TI-EjeKckr',
        'TrgInjKckr': 'SI-01SA:TI-InjDpKckr'}

    trg_propties = ('State-Sel', 'Polarity-Sel', 'Src-Sel', 'NrPulses-SP',
                    'Duration-SP', 'Delay-SP', 'Status-Mon')
    for attr, trg_name in _triggers.items():
        setattr(Const, attr, trg_name)
        for p in trg_propties:
            p = _PVName(p)
            new_attr = attr+'_'+p.propty_name
            setattr(Const, new_attr, trg_name+':'+p)

    def __init__(self, ramp_config=None, prefix=_PREFIX,
                 connection_callback=None, callback=None):
        """Init."""
        self._ramp_config = ramp_config
        properties = self._define_properties(prefix, connection_callback,
                                             callback)
        super().__init__(properties)

    def get_ramp_config(self, ramp_config):
        """Receive BoosterRamp configuration."""
        self._ramp_config = ramp_config

    # --- timing setup commands ---

    def cmd_setup(self, timeout=_TIMEOUT_DFLT):
        """Do setup TI subsystem to ramp."""
        sp_dic = self.ramp_basicsetup.copy()
        ppties = list(sp_dic.keys())
        for ppty in ppties:
            if 'Status-Mon' in ppty:
                sp_dic.pop(ppty)
        return self._command(sp_dic, timeout)

    def cmd_config_ramp(self, events_inj, events_eje, timeout=_TIMEOUT_DFLT):
        """Apply ramp_config values to TI subsystem."""
        if self._ramp_config is None:
            return

        sp_dic = dict()
        c = ConnTI.Const

        # Triggers delays
        sp_dic[c.TrgMags_Delay] = self._ramp_config.ti_params_ps_ramp_delay
        sp_dic[c.TrgCorrs_Delay] = self._ramp_config.ti_params_ps_ramp_delay
        sp_dic[c.TrgLLRFRmp_Delay] = self._ramp_config.ti_params_rf_ramp_delay

        # Event delays
        sp_dic[c.EvtRmpBO_Delay] = 0
        delays = self.calc_evts_delay(events_inj, events_eje)
        sp_dic[c.EvtLinac_Delay] = delays['Linac']
        sp_dic[c.EvtInjBO_Delay] = delays['InjBO']
        sp_dic[c.EvtInjSI_Delay] = delays['InjSI']
        sp_dic[c.TrgInjKckr_Delay] = delays[c.TrgInjKckr_Delay]
        for event in events_inj:
            attr = getattr(c, 'Evt'+event+'_Delay')
            sp_dic[attr] = delays[event]
        for event in events_eje:
            attr = getattr(c, 'Evt'+event+'_Delay')
            sp_dic[attr] = delays[event]

        # Update ramp_configsetup
        self.update_ramp_configsetup(events_inj, events_eje, delays)

        return self._command(sp_dic, timeout)

    def cmd_start_ramp(self, timeout=_TIMEOUT_DFLT):
        """Start EVG continuous events."""
        return self._command(
            {ConnTI.Const.EVG_ContinuousEvt: _TIConst.DsblEnbl.Enbl}, timeout)

    def cmd_stop_ramp(self, timeout=_TIMEOUT_DFLT):
        """Stop EVG continuous events."""
        return self._command(
            {ConnTI.Const.EVG_ContinuousEvt: _TIConst.DsblEnbl.Dsbl}, timeout)

    def cmd_start_injection(self, timeout=_TIMEOUT_DFLT):
        """Start EVG injection events."""
        return self._command(
            {ConnTI.Const.EVG_InjectionEvt: _TIConst.DsblEnbl.Enbl}, timeout)

    def cmd_stop_injection(self, timeout=_TIMEOUT_DFLT):
        """Stop EVG injection events."""
        return self._command(
            {ConnTI.Const.EVG_InjectionEvt: _TIConst.DsblEnbl.Dsbl}, timeout)

    def cmd_set_magnet_trigger_state(self, state, timeout=_TIMEOUT_DFLT):
        """Set Mag trigger state."""
        c = ConnTI.Const
        sp_dic = {c.TrgMags_State: state,
                  c.TrgCorrs_State: state}
        return self._command(sp_dic, timeout)

    def cmd_update_evts(self, timeout=_TIMEOUT_DFLT):
        """Update events."""
        return self._command(
            {ConnTI.Const.EVG_UpdateEvt: 1}, timeout)

    # --- timing mode check ---

    def check_intlk(self):
        """Check if interlock is reset."""
        return self._check({ConnTI.Const.Intlk: 0})

    def check_setup_ramp(self):
        """Check if ramp basic setup is implemented."""
        return self._check(self.ramp_basicsetup)

    def check_ramping(self):
        """Check if continuous events are enabled."""
        return self._check(
            {ConnTI.Const.EVG_ContinuousEvt: _TIConst.DsblEnbl.Enbl})

    def check_injecting(self):
        """Check if injection events are enabled."""
        return self._check(
            {ConnTI.Const.EVG_InjectionEvt: _TIConst.DsblEnbl.Enbl})

    # --- helper methods ---

    def calc_evts_delay(self, events_inj=list(), events_eje=list()):
        """Calculate event delays."""
        if self._ramp_config is None:
            return
        if not self.connected:
            return

        c = ConnTI.Const
        bo_rev = 1e6 * c.BO_HarmNum / self.get_readback(c.RFFreq)

        # Injection
        injection_time = self._ramp_config.ti_params_injection_time*1e3
        if self.get_readback(c.LinacEgun_SglBun_State):
            egun_dly = self.get_readback(c.TrgEGunSglBun_Delay)
            egun_src = self.get_readback_string(c.TrgEGunSglBun_Src)
        else:
            egun_dly = self.get_readback(c.TrgEGunMultBun_Delay)
            egun_src = self.get_readback_string(c.TrgEGunMultBun_Src)
        delay_inj = injection_time - egun_dly

        curr_dly = self.get_readback(getattr(c, 'Evt'+egun_src+'_Delay'))
        dlt_inj_dly = delay_inj - curr_dly
        dlt_inj_dly = int(dlt_inj_dly/bo_rev)*bo_rev

        # Ejection
        ejection_time = self._ramp_config.ti_params_ejection_time*1e3
        ejekckr_dly = self.get_readback(c.TrgEjeKckr_Delay)
        delay_eje = ejection_time - ejekckr_dly

        curr_injsi_dly = self.get_readback(c.EvtInjSI_Delay)
        dlt_eje_dly = delay_eje - curr_injsi_dly
        dlt_eje_dly = int(dlt_eje_dly/bo_rev)*bo_rev

        # calc delays
        delays = dict()

        events_inj = sorted(events_inj)
        if 'Linac' not in events_inj:
            events_inj.append('Linac')
        if 'InjBO' not in events_inj:
            events_inj.append('InjBO')
        for event in events_inj:
            attr = getattr(c, 'Evt'+event+'_Delay')
            curr = self.get_readback(attr)
            delays[event] = curr + dlt_inj_dly

        events_eje = sorted(events_eje)
        if 'InjSI' not in events_eje:
            events_eje.append('InjSI')
        for event in events_eje:
            attr = getattr(c, 'Evt'+event+'_Delay')
            curr = self.get_readback(attr)
            delays[event] = curr + dlt_eje_dly

        injkckr_dly = self.get_readback(c.TrgInjKckr_Delay)
        delays[c.TrgInjKckr_Delay] = injkckr_dly + dlt_eje_dly
        return delays

    def update_ramp_configsetup(self, events_inj, events_eje, delays):
        """Update ramp_configsetup dict."""
        c = ConnTI.Const
        self.ramp_configsetup.update({c.EvtRmpBO_Delay: 0})
        self.ramp_configsetup.update(
            {c.TrgMags_Delay: self._ramp_config.ti_params_ps_ramp_delay})
        self.ramp_configsetup.update(
            {c.TrgCorrs_Delay: self._ramp_config.ti_params_ps_ramp_delay})
        self.ramp_configsetup.update(
            {c.TrgLLRFRmp_Delay: self._ramp_config.ti_params_rf_ramp_delay})
        for events in [events_inj, events_eje]:
            for evt in events:
                attr = getattr(c, 'Evt'+evt+'_Delay')
                self.ramp_configsetup.update({attr: delays[evt]})

    def get_injection_time(self):
        """Return injection time."""
        c = ConnTI.Const
        # curr_linac_dly = self.get_readback(c.EvtLinac_Delay)
        curr_injbo_dly = self.get_readback(c.EvtInjBO_Delay)
        egun_dly = self.get_readback(c.TrgEGunSglBun_Delay) \
            if self.get_readback(c.LinacEgun_SglBun_State) \
            else self.get_readback(c.TrgEGunMultBun_Delay)
        return curr_injbo_dly + egun_dly

    def get_ejection_time(self):
        """Return ejection time."""
        c = ConnTI.Const
        curr_injsi_dly = self.get_readback(c.EvtInjSI_Delay)
        ejekckr_dly = self.get_readback(c.TrgEjeKckr_Delay)
        return curr_injsi_dly + ejekckr_dly

    # --- private methods ---

    def _define_properties(self, prefix, connection_callback, callback):
        c = ConnTI.Const

        mags_db = _get_trig_db(c.TrgMags)
        corrs_db = _get_trig_db(c.TrgCorrs)
        llrf_db = _get_trig_db(c.TrgLLRFRmp)

        self.ramp_basicsetup = {
            # EVG
            c.EVG_DevEnbl: _TIConst.DsblEnbl.Enbl,
            # Mags trigger
            c.TrgMags_State: _TIConst.DsblEnbl.Enbl,
            c.TrgMags_Polarity: _TIConst.TrigPol.Normal,
            c.TrgMags_Src: mags_db['Src-Sel']['enums'].index('RmpBO'),
            c.TrgMags_NrPulses: 1,
            c.TrgMags_Duration: 150.0,       # [us]
            # c.TrgMags_Status: 0,
            # Corrs trigger
            c.TrgCorrs_State: _TIConst.DsblEnbl.Enbl,
            c.TrgCorrs_Polarity: _TIConst.TrigPol.Normal,
            c.TrgCorrs_Src: corrs_db['Src-Sel']['enums'].index('RmpBO'),
            c.TrgCorrs_NrPulses: 1,
            c.TrgCorrs_Duration: 150.0,      # [us]
            # c.TrgCorrs_Status: 0,
            # LLRFRmp trigger
            c.TrgLLRFRmp_State: _TIConst.DsblEnbl.Enbl,
            c.TrgLLRFRmp_Polarity: _TIConst.TrigPol.Normal,
            c.TrgLLRFRmp_Src: llrf_db['Src-Sel']['enums'].index('RmpBO'),
            c.TrgLLRFRmp_NrPulses: 1,
            c.TrgLLRFRmp_Duration: 150.0}    # [us]
        #     c.TrgLLRFRmp_Status: 0}

        self.ramp_configsetup = {
            # Event delays
            c.EvtLinac_Delay: None,          # [us]
            c.EvtInjBO_Delay: None,          # [us]
            c.EvtRmpBO_Delay: None,          # [us]
            c.EvtInjSI_Delay: None,          # [us]
            c.EvtStudy_Delay: None,          # [us]
            c.TrgInjKckr_Delay: None,        # [us]
            # Mags trigger
            c.TrgMags_Delay: 0.0,           # [us]
            # Corrs trigger
            c.TrgCorrs_Delay: 0.0,          # [us]
            # LLRFRmp trigger
            c.TrgLLRFRmp_Delay: 0.0}        # [us]

        self._evgcontrol_propties = {
            c.EVG_ContinuousEvt: _TIConst.DsblEnbl.Dsbl,
            c.EVG_InjectionEvt: _TIConst.DsblEnbl.Dsbl,
            c.EVG_UpdateEvt: None}

        self._reading_propties = {
            # EGun trigger delays
            c.RFFreq: None,
            c.TrgEGunSglBun_Delay: 0,     # [us]
            c.TrgEGunSglBun_Src: None,
            c.TrgEGunMultBun_Delay: 0,    # [us]
            c.TrgEGunMultBun_Src: None,
            # EjeKckr trigger delay
            c.TrgEjeKckr_Delay: 0,        # [us]
            # LinacEgun Mode
            c.LinacEgun_SglBun_State: 0,
            c.LinacEgun_MultBun_State: 0,
            # Intlk
            c.Intlk: 0}

        propty2defaultvalue = self.ramp_basicsetup.copy()
        propty2defaultvalue.update(self.ramp_configsetup)
        propty2defaultvalue.update(self._evgcontrol_propties)
        propty2defaultvalue.update(self._reading_propties)

        properties = list()
        for propty, default_value in propty2defaultvalue.items():
            properties.append(
                _EpicsProperty(propty, prefix, default_value,
                               connection_callback=connection_callback,
                               callback=callback))
        return properties

    def _command(self, setpoints, timeout):
        if self.connected:
            return self.set_setpoints_check(setpoints, timeout=timeout,
                                            rel_tol=0.05, abs_tol=0.008)
        else:
            return False

    def _check(self, readbacks):
        for name, value in readbacks.items():
            if value is None:
                continue
            elif isinstance(value, float):
                if not _math.isclose(value, self.get_readback(name),
                                     abs_tol=0.008):
                    return False
            elif not self.get_readback(name) == value:
                return False
        return True


class ConnPS(_EpicsPropsList):
    """Power supplies connector class."""

    def __init__(self, ramp_config=None, prefix=_PREFIX,
                 connection_callback=None, callback=None):
        """Init."""
        self._ramp_config = ramp_config
        self._get_psnames()
        properties = self._define_properties(prefix, connection_callback,
                                             callback)
        super().__init__(properties)

    @property
    def psnames(self):
        """Return psnames."""
        return self._psnames

    def get_ramp_config(self, ramp_config):
        """Receive BoosterRamp configuration."""
        self._ramp_config = ramp_config

    # --- power supplies commands ---

    def cmd_pwrstate_on(self, timeout=_TIMEOUT_PWRSTATE_ON):
        """Turn all power supplies on."""
        return self._command_all('PwrState-Sel',
                                 _PSConst.PwrStateSel.On,
                                 desired_readback=_PSConst.PwrStateSts.On,
                                 timeout=timeout)

    def cmd_pwrstate_off(self, timeout=_TIMEOUT_PWRSTATE_OFF):
        """Turn all power supplies off."""
        return self._command_all('PwrState-Sel',
                                 _PSConst.PwrStateSel.Off,
                                 desired_readback=_PSConst.PwrStateSts.Off,
                                 timeout=timeout)

    def cmd_opmode_slowref(self, timeout=_TIMEOUT_OPMODE_CHANGE):
        """Select SlowRef opmode for all power supplies."""
        return self._command_all('OpMode-Sel',
                                 _PSConst.OpMode.SlowRef,
                                 desired_readback=_PSConst.States.SlowRef,
                                 timeout=timeout)

    def cmd_opmode_cycle(self, timeout=_TIMEOUT_OPMODE_CHANGE):
        """Select Cycle opmode for all power supplies."""
        return self._command_all('OpMode-Sel',
                                 _PSConst.OpMode.Cycle,
                                 desired_readback=_PSConst.States.Cycle,
                                 timeout=timeout)

    def cmd_opmode_rmpwfm(self, timeout=_TIMEOUT_OPMODE_CHANGE):
        """Select RmpWfm opmode for all power supplies."""
        return self._command_all('OpMode-Sel',
                                 _PSConst.OpMode.RmpWfm,
                                 desired_readback=_PSConst.States.RmpWfm,
                                 timeout=timeout)

    def cmd_wfm(self, psnames=list(), timeout=_TIMEOUT_DFLT):
        """Set wfmdata of all powersupplies."""
        if self._ramp_config is None:
            return False
        pwrsupplys = psnames if psnames else self.psnames
        sp_dic = dict()
        for psn in pwrsupplys:
            sp_dic[psn+':Wfm-SP'] = \
                self._ramp_config.ps_waveform_get_currents(psn)
        return self.set_setpoints_check(sp_dic, timeout=timeout, abs_tol=1e-5)

    # --- power supplies checks ---

    def check_pwrstate_on(self):
        """Check pwrstates of all power supplies are On."""
        return self._check_all('PwrState-Sel', _PSConst.PwrStateSts.On)

    def check_opmode_slowref(self):
        """Check opmodes of all power supplies ar SlowRef."""
        return self._check_all('OpMode-Sel', _PSConst.OpMode.SlowRef)

    def check_opmode_cycle(self):
        """Check opmodes of all power supplies ar Cycle."""
        return self._check_all('OpMode-Sel', _PSConst.OpMode.Cycle)

    def check_opmode_rmpwfm(self):
        """Check opmodes of all power supplies ar RmpWfm."""
        return self._check_all('OpMode-Sel', _PSConst.OpMode.RmpWfm)

    def check_intlksoft(self):
        """Check if software interlocks are reset."""
        return self._check_all('IntlkSoft-Mon', 0)

    def check_intlkhard(self):
        """Check if hardware interlocks are reset."""
        return self._check_all('IntlkHard-Mon', 0)

    # --- private methods ---

    def _get_psnames(self):
        self._psnames = _PSSearch.get_psnames({'sec': 'BO', 'dis': 'PS'})

    def _define_properties(self, prefix, connection_callback, callback):
        properties = []
        for psname in self._psnames:
            properties.append(
                _EpicsProperty(psname + ':PwrState-Sel', prefix,
                               connection_callback=connection_callback,
                               callback=callback))
            properties.append(
                _EpicsProperty(psname + ':OpMode-Sel', prefix,
                               connection_callback=connection_callback,
                               callback=callback))
            properties.append(
                _EpicsProperty(psname + ':Wfm-SP', prefix,
                               connection_callback=connection_callback,
                               callback=callback))
            properties.append(
                _EpicsProperty(psname + ':IntlkSoft-Mon', prefix,
                               connection_callback=connection_callback,
                               callback=callback))
            properties.append(
                _EpicsProperty(psname + ':IntlkHard-Mon', prefix,
                               connection_callback=connection_callback,
                               callback=callback))
        return properties

    def _command_all(self, prop, setpoint, desired_readback=None,
                     timeout=_TIMEOUT_DFLT):
        """Exec command for all power supplies."""
        sp_dic = dict()
        rb_dic = dict()
        for psname in self.psnames:
            name = psname + ':' + prop
            check_val = desired_readback if desired_readback else setpoint
            if not self._check_pwrsupply(psname, prop, check_val):
                sp_dic[name] = setpoint
                rb_dic[name] = check_val
        return self.set_setpoints_check(setpoints=sp_dic,
                                        desired_readbacks=rb_dic,
                                        timeout=timeout,
                                        abs_tol=1e-5)

    def _check_pwrsupply(self, psname, prop, value):
        """Check a prop of a power supplies for a value."""
        if isinstance(value, (_np.ndarray, float)):
            is_ok = _np.isclose(
                self.get_readback(psname + ':' + prop), value, atol=1e-5)
        else:
            is_ok = self.get_readback(psname + ':' + prop) == value
        if not is_ok:
            return False
        return True

    def _check_all(self, prop, value):
        """Check a prop of all power supplies for a value."""
        for psname in self.psnames:
            if not self._check_pwrsupply(psname, prop, value):
                return False
        return True


class ConnRF(_EpicsPropsList):
    """RF connector class."""

    class Const(_csdev.Const):
        """Properties names."""

        KV_2_V = 1e3
        DevName = 'BR-RF-DLLRF-01'
        Rmp_Enbl = DevName + ':RmpEnbl-Sel'
        Rmp_Ts1 = DevName + ':RmpTs1-SP'
        Rmp_Ts2 = DevName + ':RmpTs2-SP'
        Rmp_Ts3 = DevName + ':RmpTs3-SP'
        Rmp_Ts4 = DevName + ':RmpTs4-SP'
        Rmp_VoltBot = 'RA-RaBO01:RF-LLRF:RmpAmpVCavBot-SP'
        Rmp_VoltTop = 'RA-RaBO01:RF-LLRF:RmpAmpVCavTop-SP'
        Rmp_PhsBot = DevName + ':RmpPhsBot-SP'
        Rmp_PhsTop = DevName + ':RmpPhsTop-SP'
        Rmp_Intlk = DevName + ':Intlk-Mon'
        Rmp_RmpReady = DevName + ':RmpReady-Mon'
        Rmp_Times = [Rmp_Ts1, Rmp_Ts2, Rmp_Ts3, Rmp_Ts4]
        Rmp_Volts = [Rmp_VoltBot, Rmp_VoltTop]
        Rmp_Phss = [Rmp_PhsBot, Rmp_PhsTop]

    def __init__(self, ramp_config=None, prefix=_PREFIX,
                 connection_callback=None, callback=None):
        """Init."""
        self._ramp_config = ramp_config
        properties = self._define_properties(prefix, connection_callback,
                                             callback)
        super().__init__(properties)

    def get_ramp_config(self, ramp_config):
        """Receive BoosterRamp configuration."""
        self._ramp_config = ramp_config

    # --- RF commands ---

    def cmd_ramping_enable(self, timeout=_TIMEOUT_DFLT):
        """Turn RF ramping enable."""
        return self.set_setpoints_check(
            {ConnRF.Const.Rmp_Enbl: ConnRF.Const.DsblEnbl.Enbl},
            timeout=timeout)

    def cmd_ramping_disable(self, timeout=_TIMEOUT_DFLT):
        """Turn RF ramping disable."""
        return self.set_setpoints_check(
            {ConnRF.Const.Rmp_Enbl: ConnRF.Const.DsblEnbl.Dsbl},
            timeout=timeout)

    def cmd_config_ramp(self, timeout=_TIMEOUT_DFLT):
        """Apply ramp_config values to RF subsystem."""
        c = ConnRF.Const
        allconfig = self.get_propty_2_config_ramp_dict()

        # verify setpoints needed
        ppty2set = dict()
        currconfig = dict()
        for ppty, val in allconfig.items():
            currconfig[ppty] = self.get_setpoint(ppty)
            if not _np.isclose(currconfig[ppty], val, rtol=1e-3):
                ppty2set[ppty] = val

        # verify need to restart the ramp:
        # times changed or voltage ramp sign changed
        desr_dir = _np.sign(
            allconfig[c.Rmp_VoltTop] - allconfig[c.Rmp_VoltBot])
        impl_dir = _np.sign(
            currconfig[c.Rmp_VoltTop] - currconfig[c.Rmp_VoltBot])
        need_restart_ramp = set(c.Rmp_Times) & ppty2set.keys() or \
            desr_dir != impl_dir

        if need_restart_ramp:
            # disable ramp
            stsdsbl = self.cmd_ramping_disable(timeout=2)
            if not stsdsbl[0]:
                return stsdsbl

            # perform setpoints
            status = self.set_setpoints_check(
                ppty2set, rel_tol=1e-3, timeout=timeout)
            if not status[0]:
                return status

            # enable ramp
            stsenbl = self.cmd_ramping_enable(timeout=2)
            if not stsenbl[0]:
                return stsenbl
        else:
            # set phases and voltages: ensure VoltBot is set before VoltTop
            order = [c.Rmp_PhsBot, c.Rmp_VoltBot, c.Rmp_PhsTop, c.Rmp_VoltTop]
            status = self.set_setpoints_check(
                ppty2set, timeout=timeout, rel_tol=1e-3,
                order=[ppty for ppty in order if ppty in ppty2set])
        return status

    # --- RF checks ---

    def check_config_ramp(self):
        """Check if configured to ramp."""
        return self._check(self.get_propty_2_config_ramp_dict())

    def check_intlk(self):
        """Check if hardware interlocks are reset."""
        return self._check({ConnRF.Const.Rmp_Intlk: 0})

    def check_rmpready(self):
        """Check if ramp increase was concluded."""
        return self._check({ConnRF.Const.Rmp_Enbl: 1,
                            ConnRF.Const.Rmp_RmpReady: 1})

    # --- auxiliary method ---

    def get_propty_2_config_ramp_dict(self):
        """Return dict of PVs to check to be according to config."""
        dic = dict()
        c = ConnRF.Const
        dic[c.Rmp_Ts1] = self._ramp_config.rf_ramp_bottom_duration
        dic[c.Rmp_Ts2] = self._ramp_config.rf_ramp_rampup_duration
        dic[c.Rmp_Ts3] = self._ramp_config.rf_ramp_top_duration
        dic[c.Rmp_Ts4] = self._ramp_config.rf_ramp_rampdown_duration
        dic[c.Rmp_PhsBot] = self._ramp_config.rf_ramp_bottom_phase
        dic[c.Rmp_VoltBot] = self._ramp_config.rf_ramp_bottom_voltage*c.KV_2_V
        dic[c.Rmp_PhsTop] = self._ramp_config.rf_ramp_top_phase
        dic[c.Rmp_VoltTop] = self._ramp_config.rf_ramp_top_voltage*c.KV_2_V
        return dic

    # --- private methods ---

    def _define_properties(self, prefix, connection_callback, callback):
        c = ConnRF.Const
        propty2defaultvalue = {
            c.Rmp_Enbl: ConnRF.Const.DsblEnbl.Enbl,
            c.Rmp_Ts1: _rutil.DEFAULT_RF_RAMP_BOTTOM_DURATION,
            c.Rmp_Ts2: _rutil.DEFAULT_RF_RAMP_RAMPUP_DURATION,
            c.Rmp_Ts3: _rutil.DEFAULT_RF_RAMP_TOP_DURATION,
            c.Rmp_Ts4: _rutil.DEFAULT_RF_RAMP_RAMPDOWN_DURATION,
            c.Rmp_VoltBot: _rutil.DEFAULT_RF_RAMP_BOTTOM_VOLTAGE,
            c.Rmp_VoltTop: _rutil.DEFAULT_RF_RAMP_TOP_VOLTAGE,
            c.Rmp_PhsBot: _rutil.DEFAULT_RF_RAMP_BOTTOM_PHASE,
            c.Rmp_PhsTop: _rutil.DEFAULT_RF_RAMP_TOP_PHASE,
            c.Rmp_Intlk: None,
            c.Rmp_RmpReady: None}

        properties = list()
        for propty, default_value in propty2defaultvalue.items():
            properties.append(
                _EpicsProperty(propty, prefix, default_value,
                               connection_callback=connection_callback,
                               callback=callback))
        return properties

    def _check(self, readbacks):
        for name, value in readbacks.items():
            if value is None:
                continue
            if not self.get_readback(name) == value:
                return False
        return True


class ConnSOFB(_EpicsPropsList):
    """SOFB connector class."""

    IOC_PREFIX = 'BO-Glob:AP-SOFB'

    def __init__(self, prefix=_PREFIX,
                 connection_callback=None, callback=None):
        """Init."""
        properties = self._define_properties(prefix, connection_callback,
                                             callback)
        super().__init__(properties)

    def get_deltakicks(self):
        """Get CH and CV delta kicks calculated by SOFB."""
        bo_sofb_db = _SOFBRings(acc='BO')
        rb_dic = self.readbacks
        ch_dkicks = rb_dic[ConnSOFB.IOC_PREFIX + ':DeltaKickCH-Mon']
        ch_names = bo_sofb_db.ch_names
        cv_dkicks = rb_dic[ConnSOFB.IOC_PREFIX + ':DeltaKickCV-Mon']
        cv_names = bo_sofb_db.cv_names

        corrs2dkicks_dict = dict()
        for idx, chn in enumerate(ch_names):
            corrs2dkicks_dict[chn] = ch_dkicks[idx]
        for idx, cvn in enumerate(cv_names):
            corrs2dkicks_dict[cvn] = cv_dkicks[idx]
        return corrs2dkicks_dict

    def _define_properties(self, prefix, connection_callback, callback):
        properties = list()
        for ppty in ['DeltaKickCH-Mon', 'DeltaKickCV-Mon']:
            properties.append(
                _EpicsProperty(ConnSOFB.IOC_PREFIX + ':' + ppty, prefix,
                               connection_callback=connection_callback,
                               callback=callback))
        return properties



#!/usr/bin/env python3.6

"""Module to test opticscorr."""

import unittest
import numpy as np
from siriuspy import util
from siriuspy.opticscorr.opticscorr import OpticsCorr

PUB_INTERFACE = (
    'magnetfams_ordering',
    'nominal_matrix',
    'nominal_intstrengths',
    'nominal_opticsparam',
    'magnetfams_focusing',
    'magnetfams_defocusing',
    'matrix_add_svd',
    'matrix_add_2knobs',
    'matrix_prop_svd',
    'matrix_prop_2knobs',
    'inverse_matrix_add_svd',
    'inverse_matrix_add_2knobs',
    'inverse_matrix_prop_svd',
    'inverse_matrix_prop_2knobs',
    'calculate_delta_intstrengths',
    'calculate_opticsparam'
)


class TestOpticsCorr(unittest.TestCase):
    """Test OpticsCorr class."""

    def setUp(self):
        """Setup tests."""
        # Attributs fot simulating normal operation
        self.magnetfams_ordering_ok = (
            'QFA', 'QFB', 'QFP', 'QDA', 'QDB1', 'QDB2', 'QDP1', 'QDP2')
        self.magnetfams_focusing_ok = ('QFA', 'QFB', 'QFP')
        self.magnetfams_defocusing_ok = (
            'QDA', 'QDB1', 'QDB2', 'QDP1', 'QDP2')
        self.nominal_matrix_ok = [2.7280, 8.5894, 4.2995, 0.5377,
                                  1.0906, 2.0004, 0.5460, 1.0012,
                                  -1.3651, -3.5532, -1.7657, -2.3652,
                                  -4.7518, -1.9781, -2.3601, -0.9839]
        self.nominal_intstrengths_ok = [0.7146, 1.2344, 1.2344, -0.2270,
                                        -0.2809, -0.4783, -0.2809, -0.4783]
        self.nominal_opticsparam_ok = [0.0, 0.0]

        # Attributes to simulate type errors
        self.magnetfams_ordering_error1 = (1, 2, 3, 4)
        self.magnetfams_focusing_error1 = (1, 2)
        self.magnetfams_defocusing_error1 = (3, 4)
        self.nominal_matrix_error1 = [
            '2.7280', '8.5894', '0.5460', '1.0012',
            '-1.3651', '-3.5532', '-2.3601', '-0.9839']
        self.nominal_intstrengths_error1 = [
            '0.7146', '1.2344', '-0.2809', '-0.4783']
        self.nominal_opticsparam_error1 = ['0.0', '0.0']

        # Attributes to simulate matrix invertion errors
        self.magnetfams_ordering_error2 = ('QF', 'QD')
        self.magnetfams_focusing_error2 = ('QF')
        self.magnetfams_defocusing_error2 = ('QD')
        self.nominal_matrix_error2 = [0.0, 2, 0.1, 4]
        self.nominal_intstrengths_error2 = [0.3, -1.5]

        # Attributes to simulate value errors
        self.magnetfams_ordering_error3 = ()
        self.magnetfams_focusing_error3 = ()
        self.magnetfams_defocusing_error3 = ()
        self.nominal_matrix_error3 = [2.7280, 8.5894, 4.2995, 0.5377,
                                      1.0906, 2.0004, 0.5460, 1.0012,
                                      -1.3651, -3.5532, -1.7657, -2.3652,
                                      -4.7518, -1.9781, -2.3601]
        self.nominal_opticsparam_error3 = [0.0, 0.0, 0.0]

        # Attributes to test correction with some families, not all
        self.magnetfams_focusing_somefams = ('QFA', 'QFB')
        self.magnetfams_defocusing_somefams = ('QDA', 'QDB2', 'QDB1')

    def test_public_interface(self):
        """Test module's public interface."""
        valid = util.check_public_interface_namespace(
            OpticsCorr, PUB_INTERFACE, print_flag=True)
        self.assertTrue(valid)

    def test_type_errors(self):
        """Test raise type errors."""
        with self.assertRaises(TypeError):
            self.opticscorr = OpticsCorr(
                self.magnetfams_ordering_error1,
                self.nominal_matrix_ok,
                self.nominal_intstrengths_ok,
                self.nominal_opticsparam_ok,
                self.magnetfams_focusing_ok,
                self.magnetfams_defocusing_ok)

        with self.assertRaises(TypeError):
            self.opticscorr = OpticsCorr(
                self.magnetfams_ordering_ok,
                self.nominal_matrix_ok,
                self.nominal_intstrengths_ok,
                self.nominal_opticsparam_ok,
                self.magnetfams_focusing_error1,
                self.magnetfams_defocusing_ok)

        with self.assertRaises(TypeError):
            self.opticscorr = OpticsCorr(
                self.magnetfams_ordering_ok,
                self.nominal_matrix_ok,
                self.nominal_intstrengths_ok,
                self.nominal_opticsparam_ok,
                self.magnetfams_focusing_ok,
                self.magnetfams_defocusing_error1)

        with self.assertRaises(TypeError):
            self.opticscorr = OpticsCorr(
                self.magnetfams_ordering_ok,
                self.nominal_matrix_error1,
                self.nominal_intstrengths_ok,
                self.nominal_opticsparam_ok,
                self.magnetfams_focusing_ok,
                self.magnetfams_defocusing_ok)

        with self.assertRaises(TypeError):
            self.opticscorr = OpticsCorr(
                self.magnetfams_ordering_ok,
                self.nominal_matrix_ok,
                self.nominal_intstrengths_error1,
                self.nominal_opticsparam_ok,
                self.magnetfams_focusing_ok,
                self.magnetfams_defocusing_ok)

        with self.assertRaises(TypeError):
            self.opticscorr = OpticsCorr(
                self.magnetfams_ordering_ok,
                self.nominal_matrix_ok,
                self.nominal_intstrengths_ok,
                self.nominal_opticsparam_error1,
                self.magnetfams_focusing_ok,
                self.magnetfams_defocusing_ok)

    def test_matrix_invertion_errors(self):
        """Test raise exceptions on matrix invertion errors."""
        with self.assertRaises(Exception):
            self.opticscorr = OpticsCorr(
                self.magnetfams_ordering_error2,
                self.nominal_matrix_error2,
                self.nominal_intstrengths_error2,
                self.nominal_opticsparam_ok,
                self.magnetfams_focusing_error2,
                self.magnetfams_defocusing_error2)

    def test_value_errors(self):
        """Test raise value errors."""
        with self.assertRaises(ValueError):
            self.opticscorr = OpticsCorr(
                self.magnetfams_ordering_error3,
                self.nominal_matrix_ok,
                self.nominal_intstrengths_ok,
                self.nominal_opticsparam_ok,
                self.magnetfams_focusing_ok,
                self.magnetfams_defocusing_ok)

        with self.assertRaises(ValueError):
            self.opticscorr = OpticsCorr(
                self.magnetfams_ordering_ok,
                self.nominal_matrix_error3,
                self.nominal_intstrengths_ok,
                self.nominal_opticsparam_ok,
                self.magnetfams_focusing_ok,
                self.magnetfams_defocusing_ok)

        with self.assertRaises(ValueError):
            self.opticscorr = OpticsCorr(
                self.magnetfams_ordering_ok,
                self.nominal_matrix_ok,
                self.nominal_intstrengths_ok,
                self.nominal_opticsparam_error3,
                self.magnetfams_focusing_ok,
                self.magnetfams_defocusing_ok)

        with self.assertRaises(ValueError):
            self.opticscorr = OpticsCorr(
                self.magnetfams_ordering_ok,
                self.nominal_matrix_ok,
                self.nominal_intstrengths_ok,
                self.nominal_opticsparam_ok,
                self.magnetfams_focusing_error3,
                self.magnetfams_defocusing_ok)

        with self.assertRaises(ValueError):
            self.opticscorr = OpticsCorr(
                self.magnetfams_ordering_ok,
                self.nominal_matrix_ok,
                self.nominal_intstrengths_ok,
                self.nominal_opticsparam_ok,
                self.magnetfams_focusing_ok,
                self.magnetfams_defocusing_error3)

    def test_magnetfams_ordering(self):
        """Test magnetfams_ordering property."""
        self.opticscorr = OpticsCorr(self.magnetfams_ordering_ok,
                                     self.nominal_matrix_ok,
                                     self.nominal_intstrengths_ok,
                                     self.nominal_opticsparam_ok,
                                     self.magnetfams_focusing_ok,
                                     self.magnetfams_defocusing_ok)

        propty = self.opticscorr.magnetfams_ordering
        self.assertIsInstance(propty, tuple)
        for item in propty:
            self.assertIsInstance(item, str)
        self.assertGreaterEqual(len(propty), 1)

    def test_magnetfams_focusing(self):
        """Test test_magnetfams_focusing property."""
        self.opticscorr = OpticsCorr(self.magnetfams_ordering_ok,
                                     self.nominal_matrix_ok,
                                     self.nominal_intstrengths_ok,
                                     self.nominal_opticsparam_ok,
                                     self.magnetfams_focusing_ok,
                                     self.magnetfams_defocusing_ok)

        propty = self.opticscorr.magnetfams_focusing
        self.assertIsInstance(propty, tuple)
        for item in propty:
            self.assertIsInstance(item, str)
        self.assertGreaterEqual(len(propty), 1)

    def test_magnetfams_defocusing(self):
        """Test test_magnetfams_defocusing property."""
        self.opticscorr = OpticsCorr(self.magnetfams_ordering_ok,
                                     self.nominal_matrix_ok,
                                     self.nominal_intstrengths_ok,
                                     self.nominal_opticsparam_ok,
                                     self.magnetfams_focusing_ok,
                                     self.magnetfams_defocusing_ok)

        propty = self.opticscorr.magnetfams_defocusing
        self.assertIsInstance(propty, tuple)
        for item in propty:
            self.assertIsInstance(item, str)
        self.assertGreaterEqual(len(propty), 1)

    def test_nominal_matrix(self):
        """Test nominal_matrix property."""
        self.opticscorr = OpticsCorr(self.magnetfams_ordering_ok,
                                     self.nominal_matrix_ok,
                                     self.nominal_intstrengths_ok,
                                     self.nominal_opticsparam_ok,
                                     self.magnetfams_focusing_ok,
                                     self.magnetfams_defocusing_ok)

        propty = self.opticscorr.nominal_matrix
        self.assertIsInstance(propty, np.ndarray)
        self.assertEqual(propty.shape[0], 2)
        self.assertEqual(propty.shape[1],
                         len(self.opticscorr.magnetfams_ordering))

    def test_nominal_intstrengths(self):
        """Test nominal_strengths property."""
        self.opticscorr = OpticsCorr(self.magnetfams_ordering_ok,
                                     self.nominal_matrix_ok,
                                     self.nominal_intstrengths_ok,
                                     self.nominal_opticsparam_ok,
                                     self.magnetfams_focusing_ok,
                                     self.magnetfams_defocusing_ok)

        propty = self.opticscorr.nominal_intstrengths
        self.assertIsInstance(propty, np.ndarray)
        self.assertEqual(propty.shape[0], 1)
        self.assertEqual(propty.shape[1],
                         len(self.opticscorr.magnetfams_ordering))

    def test_nominal_opticsparam(self):
        """Test nominal_opticsparam property."""
        self.opticscorr = OpticsCorr(self.magnetfams_ordering_ok,
                                     self.nominal_matrix_ok,
                                     self.nominal_intstrengths_ok,
                                     self.nominal_opticsparam_ok,
                                     self.magnetfams_focusing_ok,
                                     self.magnetfams_defocusing_ok)

        propty = self.opticscorr.nominal_opticsparam
        self.assertIsInstance(propty, np.ndarray)
        self.assertTrue(propty.shape, (2, 1))

    def test_matrix_add_svd(self):
        """Test matrix_add_svd property."""
        self.opticscorr = OpticsCorr(self.magnetfams_ordering_ok,
                                     self.nominal_matrix_ok,
                                     self.nominal_intstrengths_ok,
                                     self.nominal_opticsparam_ok,
                                     self.magnetfams_focusing_ok,
                                     self.magnetfams_defocusing_ok)

        propty = self.opticscorr.matrix_add_svd
        self.assertIsInstance(propty, np.ndarray)
        self.assertEqual(propty.shape[0], 2)
        self.assertEqual(propty.shape[1],
                         len(self.opticscorr.magnetfams_focusing) +
                         len(self.opticscorr.magnetfams_defocusing))

    def test_matrix_prop_svd(self):
        """Test matrix_prop_svd property."""
        self.opticscorr = OpticsCorr(self.magnetfams_ordering_ok,
                                     self.nominal_matrix_ok,
                                     self.nominal_intstrengths_ok,
                                     self.nominal_opticsparam_ok,
                                     self.magnetfams_focusing_ok,
                                     self.magnetfams_defocusing_ok)

        propty = self.opticscorr.matrix_prop_svd
        self.assertIsInstance(propty, np.ndarray)
        self.assertEqual(propty.shape[0], 2)
        self.assertEqual(propty.shape[1],
                         len(self.opticscorr.magnetfams_focusing) +
                         len(self.opticscorr.magnetfams_defocusing))

    def test_matrix_add_2knobs(self):
        """Test matrix_add_2knobs property."""
        self.opticscorr = OpticsCorr(self.magnetfams_ordering_ok,
                                     self.nominal_matrix_ok,
                                     self.nominal_intstrengths_ok,
                                     self.nominal_opticsparam_ok,
                                     self.magnetfams_focusing_ok,
                                     self.magnetfams_defocusing_ok)

        propty = self.opticscorr.matrix_add_2knobs
        self.assertIsInstance(propty, np.ndarray)
        self.assertEqual(propty.shape[0], 2)
        self.assertEqual(propty.shape[1], 2)

    def test_matrix_prop_2knobs(self):
        """Test matrix_prop_2knobs property."""
        self.opticscorr = OpticsCorr(self.magnetfams_ordering_ok,
                                     self.nominal_matrix_ok,
                                     self.nominal_intstrengths_ok,
                                     self.nominal_opticsparam_ok,
                                     self.magnetfams_focusing_ok,
                                     self.magnetfams_defocusing_ok)

        propty = self.opticscorr.matrix_prop_2knobs
        self.assertIsInstance(propty, np.ndarray)
        self.assertEqual(propty.shape[0], 2)
        self.assertEqual(propty.shape[1], 2)

    def test_inv_matrix_add_svd(self):
        """Test inverse_matrix_add_svd property."""
        self.opticscorr = OpticsCorr(self.magnetfams_ordering_ok,
                                     self.nominal_matrix_ok,
                                     self.nominal_intstrengths_ok,
                                     self.nominal_opticsparam_ok,
                                     self.magnetfams_focusing_ok,
                                     self.magnetfams_defocusing_ok)

        propty = self.opticscorr.inverse_matrix_add_svd
        self.assertIsInstance(propty, np.ndarray)
        self.assertEqual(propty.shape[0],
                         len(self.opticscorr.magnetfams_focusing) +
                         len(self.opticscorr.magnetfams_defocusing))
        self.assertEqual(propty.shape[1], 2)
        umat, smat, vmat = np.linalg.svd(self.opticscorr.matrix_add_svd,
                                         full_matrices=False)
        self.assertListEqual(
            list(propty.flatten()),
            list(np.dot(np.dot(vmat.T, np.diag(1/smat)), umat.T).flatten()))

    def test_inv_matrix_prop_svd(self):
        """Test inverse_matrix_prop_svd property."""
        self.opticscorr = OpticsCorr(self.magnetfams_ordering_ok,
                                     self.nominal_matrix_ok,
                                     self.nominal_intstrengths_ok,
                                     self.nominal_opticsparam_ok,
                                     self.magnetfams_focusing_ok,
                                     self.magnetfams_defocusing_ok)

        propty = self.opticscorr.inverse_matrix_prop_svd
        self.assertIsInstance(propty, np.ndarray)
        self.assertEqual(propty.shape[0],
                         len(self.opticscorr.magnetfams_focusing) +
                         len(self.opticscorr.magnetfams_defocusing))
        self.assertEqual(propty.shape[1], 2)
        umat, smat, vmat = np.linalg.svd(self.opticscorr.matrix_prop_svd,
                                         full_matrices=False)
        self.assertListEqual(
            list(propty.flatten()),
            list(np.dot(np.dot(vmat.T, np.diag(1/smat)), umat.T).flatten()))

    def test_inv_matrix_add_2knobs(self):
        """Test inverse_matrix_add_2knobs property."""
        self.opticscorr = OpticsCorr(self.magnetfams_ordering_ok,
                                     self.nominal_matrix_ok,
                                     self.nominal_intstrengths_ok,
                                     self.nominal_opticsparam_ok,
                                     self.magnetfams_focusing_ok,
                                     self.magnetfams_defocusing_ok)

        propty = self.opticscorr.inverse_matrix_add_2knobs
        self.assertIsInstance(propty, np.ndarray)
        self.assertEqual(propty.shape[0], 2)
        self.assertEqual(propty.shape[1], 2)
        umat, smat, vmat = np.linalg.svd(self.opticscorr.matrix_add_2knobs,
                                         full_matrices=False)
        self.assertListEqual(
            list(propty.flatten()),
            list(np.dot(np.dot(vmat.T, np.diag(1/smat)), umat.T).flatten()))

    def test_inv_matrix_prop_2knobs(self):
        """Test inverse_matrix_prop_2knobs property."""
        self.opticscorr = OpticsCorr(self.magnetfams_ordering_ok,
                                     self.nominal_matrix_ok,
                                     self.nominal_intstrengths_ok,
                                     self.nominal_opticsparam_ok,
                                     self.magnetfams_focusing_ok,
                                     self.magnetfams_defocusing_ok)

        propty = self.opticscorr.inverse_matrix_prop_2knobs
        self.assertIsInstance(propty, np.ndarray)
        self.assertEqual(propty.shape[0], 2)
        self.assertEqual(propty.shape[1], 2)
        umat, smat, vmat = np.linalg.svd(self.opticscorr.matrix_prop_2knobs,
                                         full_matrices=False)
        self.assertListEqual(
            list(propty.flatten()),
            list(np.dot(np.dot(vmat.T, np.diag(1/smat)), umat.T).flatten()))

    def test_calc_dintstr_allfams(self):
        """Test calculate_delta_intstrengths function (all families)."""
        self.opticscorr = OpticsCorr(self.magnetfams_ordering_ok,
                                     self.nominal_matrix_ok,
                                     self.nominal_intstrengths_ok,
                                     self.nominal_opticsparam_ok,
                                     self.magnetfams_focusing_ok,
                                     self.magnetfams_defocusing_ok)

        delta_opticsparam = [0.01, 0.01]

        # additional method
        delta_intstrengths = self.opticscorr.calculate_delta_intstrengths(
            method=1, grouping='svd', delta_opticsparam=delta_opticsparam)
        expected = [0.00027307, 0.00125755, 0.00063636, -0.00106541,
                    -0.00213666, -0.00032148, -0.00105963, -0.00015762]
        for idx, data in enumerate(delta_intstrengths):
            self.assertAlmostEqual(data, expected[idx])

        delta_intstrengths = self.opticscorr.calculate_delta_intstrengths(
            method=1, grouping='2knobs', delta_opticsparam=delta_opticsparam)
        expected = [0.00110325, 0.00110325, 0.00110325, -0.00139674,
                    -0.00139674, -0.00139674, -0.00139674, -0.00139674]
        for idx, data in enumerate(delta_intstrengths):
            self.assertAlmostEqual(data, expected[idx])

        # proportional method
        delta_intstrengths = self.opticscorr.calculate_delta_intstrengths(
            method=0, grouping='svd', delta_opticsparam=delta_opticsparam)
        expected = [-0.00052788, 0.00164248, 0.00093635, -0.0006387,
                    -0.00196284, -0.00147348, -0.00097403, -0.00072931]
        for idx, data in enumerate(delta_intstrengths):
            self.assertAlmostEqual(data, expected[idx])

        delta_intstrengths = self.opticscorr.calculate_delta_intstrengths(
            method=0, grouping='2knobs', delta_opticsparam=delta_opticsparam)
        expected = [0.00077053, 0.00133102, 0.00133102, -0.00104162,
                    -0.00128895, -0.00219475, -0.00128895, -0.00219475]
        for idx, data in enumerate(delta_intstrengths):
            self.assertAlmostEqual(data, expected[idx])

    def test_calc_dintstr_somefams(self):
        """Test calculate_delta_intstrengths function (5 families)."""
        self.opticscorr = OpticsCorr(self.magnetfams_ordering_ok,
                                     self.nominal_matrix_ok,
                                     self.nominal_intstrengths_ok,
                                     self.nominal_opticsparam_ok,
                                     self.magnetfams_focusing_somefams,
                                     self.magnetfams_defocusing_somefams)

        delta_opticsparam = [0.01, 0.01]

        delta_intstrengths = self.opticscorr.calculate_delta_intstrengths(
            method=1, grouping='svd', delta_opticsparam=delta_opticsparam)
        expected = [0.00034033, 0.0015503, 0.,
                    -0.00127996, -0.00256688, -0.00037835, 0., 0.]
        for idx, data in enumerate(delta_intstrengths):
            self.assertAlmostEqual(data, expected[idx])

    def test_calc_opticsparam(self):
        """Test calculate_opticsparam function."""
        self.opticscorr = OpticsCorr(self.magnetfams_ordering_ok,
                                     self.nominal_matrix_ok,
                                     self.nominal_intstrengths_ok,
                                     self.nominal_opticsparam_ok,
                                     self.magnetfams_focusing_ok,
                                     self.magnetfams_defocusing_ok)

        delta_intstrengths = [5.23801198e-04, 1.91863308e-03,
                              9.65048343e-04, -6.54746947e-04,
                              -1.31156341e-03, 3.07649982e-05,
                              -6.49807134e-04, 1.76177711e-05]
        opticsparam = self.opticscorr.calculate_opticsparam(delta_intstrengths)
        expected = [0.02, 0]
        for idx, data in enumerate(opticsparam):
            self.assertAlmostEqual(data, expected[idx])


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python-sirius
"""Magnet class test module."""

from unittest import TestCase
from siriuspy import util
from siriuspy.ramp import magnet


public_interface = (
    'get_magnet',
    'Magnet',
)


class TestModule(TestCase):
    """Test module interface."""

    def test_public_interface(self):
        """Test module's public interface."""
        valid = util.check_public_interface_namespace(
                magnet,
                public_interface)
        self.assertTrue(valid)


class TestMagnet(TestCase):
    """TestMagnet."""

    class Modules:
        """Modules class."""

        pass

    # ps_manames = _search.MASearch.get_pwrsupply_manames()
    # conv_maname_2_splims = _search.MASearch.conv_maname_2_splims

    data = {
        'SI-Fam:MA-B1B2':
            {'DRVH': 405.0,
             'DRVL': 0.0,
             'HIGH': 405.0,
             'HIHI': 405.0,
             'HOPR': 400.0,
             'LOLO': 0.0,
             'LOPR': 0.0,
             'LOW': 0.0,
             'TSTR': 0.5,
             'TSTV': 394.0},
        'SI-Fam:MA-QDA': {},
        'SI-Fam:MA-QDB1': {},
        'SI-Fam:MA-QFA': {},
        'SI-Fam:MA-Q1': {},
        'SI-Fam:MA-QFB': {},
        'SI-Fam:MA-QFP': {},
        'SI-Fam:MA-SFA0': {},
        'SI-Fam:MA-SFP1': {},
        'SI-Fam:MA-SDA0': {},
        'SI-Fam:MA-SDP0': {},
        'TS-Fam:MA-B': {},
        'BO-Fam:MA-B': {},
        'BO-Fam:MA-QD': {},
        'BO-Fam:MA-QF': {},
        'BO-Fam:MA-SD': {},
        'BO-Fam:MA-SF': {},
        'TB-Fam:MA-B': {},
        'SI-01M1:MA-QDA': {},
        'SI-17C1:MA-Q2': {},
        'SI-07C2:MA-CH': {},
        'SI-01C1:MA-CV': {},
        'SI-01C2:MA-CV-1': {},
        'SI-05C1:MA-QS': {},
        'SI-09M1:MA-FCH': {},
        'SI-04C3:MA-FCV': {},
        'BO-02D:MA-QS': {},
        'BO-29U:MA-CH': {},
        'BO-07U:MA-CV': {},
        'TB-01:MA-QD1': {},
        'TB-02:MA-QD2A': {},
        'TB-01:MA-CH-1': {},
        'TB-03:MA-CH': {},
        'TB-01:MA-CV-1': {},
        'TS-01:MA-QF1A': {},
        'TS-02:MA-QD2': {},
        'TS-02:MA-QF2': {},
        'TS-04:MA-CH': {},
        'TS-01:MA-CV-1': {},
        'TS-04:PM-InjSeptG-1': {},
        'TS-04:PM-InjSeptG-2': {},
        'TS-04:PM-InjSeptF': {},
        'TS-01:PM-EjeSeptG': {},
        'TS-01:PM-EjeSeptF': {},
        'BO-01D:PM-InjKckr': {},
        'BO-48D:PM-EjeKckr': {},
        'TB-04:PM-InjSept': {},
        'SI-01SA:PM-InjDpKckr': {},
        'SI-01SA:PM-InjNLKckr': {},
        'SI-01SA:PM-PingH': {},
        'SI-19C4:PM-PingV': {},
    }

    def setUp(self):
        """Setup method."""
        import siriuspy.ramp.magnet as _mod
        self.modules = TestMagnet.Modules()
        self.modules.magnet = _mod

        # search_patcher = mock.patch(
        #     'siriuspy.ramp.magnet._MAData', autospec=True)
        # self.addCleanup(search_patcher.stop)
        # self.mock = search_patcher.start()
        # self.mock.return_value.magfunc.return_value = 'dipole'
        # self.mock.conv_maname_2_splims.return_value = \
        #     {'DRVH': 405.0,
        #      'DRVL': 0.0,
        #      'HIGH': 405.0,
        #      'HIHI': 405.0,
        #      'HOPR': 400.0,
        #      'LOLO': 0.0,
        #      'LOPR': 0.0,
        #      'LOW': 0.0,
        #      'TSTR': 0.5,
        #      'TSTV': 394.0}
        # self.mock.get_splims_unit.return_value = ['A', 'Ampere']

    public_interface = (
        'maname',
        'section',
        'dipole_name',
        'family_name',
        'magfunc',
        'strength_label',
        'strength_units',
        'current_min',
        'current_max',
        'splims',
        'conv_current_2_strength',
        'conv_strength_2_current'
    )

    def test_public_interface(self):
        """Test module's public interface."""
        valid = util.check_public_interface_namespace(
                magnet.Magnet,
                TestMagnet.public_interface)
        self.assertTrue(valid)

    def _test_constructor(self):
        """Test create magnets.

        - Create a magnet for each maname in MASearchs.get_manames().
        """
        # TODO: use mocking!!!
        for maname, data in TestMagnet.data.items():
            magnet = self.modules.magnet.Magnet(maname=maname)
            self.assertEqual(magnet.maname, maname)
            self.assertIn(magnet.dipole_name, (None,
                                               'SI-Fam:MA-B1B2',
                                               'BO-Fam:MA-B',
                                               'TB-Fam:MA-B',
                                               'TS-Fam:MA-B'))
            # if data:
            #     splims = data
            #     self.assertEqual(magnet.current_max, splims['DRVH'])
            #     self.assertEqual(magnet.current_min, splims['DRVL'])

    def _test_conversions(self):
        """Test conversion current 2 strength."""
        # TODO: use mocking!!!
        for maname in TestMagnet.manames:
            if 'PM-' in maname:
                print('test_conversions: skipping {} ...'.format(maname))
                continue
            magnet = self.modules.magnet.Magnet(maname=maname)
            splims = TestMagnet.conv_maname_2_splims(maname)
            currents1 = list(splims.values())
            strengths = magnet.conv_current_2_strength(
                currents1,
                strengths_dipole=3.0,
                strengths_family=1.0)
            currents2 = magnet.conv_strength_2_current(
                strengths,
                strengths_dipole=3.0,
                strengths_family=1.0)
            for i in range(len(currents1)):
                self.assertAlmostEqual(currents1[i], currents2[i])
